from __future__ import annotations

from uuid import UUID

from cimgraph.databases import ConnectionParameters


def get_triple_cypher(obj:object, attribute:str, connection_params: ConnectionParameters) -> str:

    """
    Generates cypher query string to find the type of an object from its uri
    Args:
        mrid (str): The mRID or uri of the  object
        connection_params (ConnectionParameters): Database connection parameters
    Returns:
        query_message: query string that can be used in Neo4J
    """

    if int(connection_params.iec61970_301) > 7:
        split = 'urn:uuid:'
    else:
        split = f'{connection_params.url}#'

    query_message = f'''MATCH (n:{obj.__class__.__name__})
WHERE n.uri = "{split+obj.uri()}"
OPTIONAL MATCH (n) - [:`{attribute}`] - (m)
RETURN DISTINCT
n.uri as identifier,
"{attribute}" as attribute,
COALESCE(n.`{attribute}`,m.uri) as edge_id,
labels(m)[1] as edge_class'''
    # '{"@id":"' + COALESCE(m.uri ,"") + '","@type":"' + COALESCE((labels(m))[1],"") + '"}' as edge
    return query_message



from __future__ import annotations

import importlib
import json
import logging
from dataclasses import dataclass, field

from cimgraph.models.graph_model import GraphModel

_log = logging.getLogger(__name__)


@dataclass
class BusBranchModel(GraphModel):

    distributed_hierarchy: list[type] = field(default_factory=list)

    def __post_init__(self):
        self.cim_profile = self.connection.cim_profile
        self.cim = importlib.import_module('cimgraph.data_profile.' + self.cim_profile)

        if self.connection is not None:
            if self.distributed:
                self.initialize_distributed_model(self.container)
            else:
                self.initialize_centralized_model(self.container)
        else:
            _log.error('A ConnectionInterface must be specified')

    def initialize_centralized_model(self, container) -> None:
        self.graph = self.connection.create_new_graph(container)

    def initialize_distributed_model(self, container) -> None:
        pass
        # if len(self.distributed_hierarchy) > 0:
        #     for container_class in self.distributed_hierarchy:
        #         container_type = container_class.__class__.__name__
        #         setattr(self, container_type + 's', [])
        #         #TODO: create subclasses based on pre-defined topology
        # else:
        #     centralized_graph = self.connection.create_new_graph(container)
        #     self.get_all_edges(self.cim.PowerTransformer, centralized_graph)
        #     self.get_all_edges(self.cim.TransformerTank, centralized_graph)
        #     self.get_all_edges(self.cim.BaseVoltage, centralized_graph)


#             self.linknet = LinkNet(self.cim_profile, centralized_graph)
#             self.linknet.build_linknet([self.cim.ACLineSegment])



import json
import logging
from uuid import UUID

_log = logging.getLogger(__name__)

jsonld = dict['@id':str(UUID),'@type':str(type)]
Graph = dict[type, dict[UUID, object]]



# def jsonld_to_obj(cim:cim, json_ld:jsonld|str[jsonld]) -> object:

#     if type(json_ld) == str:
#         json_ld = json.loads(json_ld)
#     elif type(json_ld) == dict:
#         pass
#     else:
#         raise TypeError('json_ld input must be string or dict')



def create_object(class_type:type, uri:str, graph:Graph = None) -> object:
    """
    Method for creating new objects and adding them to the graph
    Required Args:
        graph: an LPG graph from a GraphModel object
        class_type: a dataclass type, such as cim.ACLineSegment
        uri: the RDF ID or mRID of the object
    Returns:
        obj: a dataclass instance with the correct identifier
    """
    # Convert uri string to a uuid
    try:
        identifier = UUID(uri.strip('_').lower())
    except:
        _log.warning(f'URI {uri} for object {class_type.__name__} is not a valid UUID')
        identifier = uri

    if graph is None:
        graph = {}

    # Add class type to graph keys if not there
    if class_type not in graph:
        graph[class_type] = {}

    # Check if object exists in graph
    if identifier in graph[class_type]:
        obj = graph[class_type][identifier]

    # If not there, create a new object and add to graph
    else:
        obj = class_type()
        obj.uuid(uri = uri)
        graph[class_type][identifier] = obj

    return obj



from __future__ import annotations

import logging

from cimgraph.databases import ConnectionParameters

_log = logging.getLogger(__name__)



def get_all_nodes_from_container(container: object, connection_params: ConnectionParameters) -> str:
    """
    Generates SPARQL query string for all nodes, terminals, and conducting equipment
    Args:

    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """
    container_class = container.__class__.__name__
    try:
        container_uri = container.uri()
    except:
        container_uri = container.mRID

    if int(connection_params.iec61970_301) > 7:
        split = 'urn:uuid:'
    else:
        split = f'{connection_params.url}#'


    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <%s>""" % connection_params.namespace
    query_message += """
        SELECT DISTINCT ?ConnectivityNode ?Terminal ?Equipment
        WHERE {
          """ #
    query_message += """
        VALUES ?identifier {"%s"}
        bind(iri(concat("%s", ?identifier)) as ?c).
        """ % (container_uri, split)

    # get ConnectivityNode objects associated with Container
    query_message += '''     {

        ?node cim:ConnectivityNode.ConnectivityNodeContainer ?c.
        ?t cim:Terminal.ConnectivityNode ?node.
        ?t cim:Terminal.ConductingEquipment ?eq.
        ?eq a ?eq_cls.

        bind(strafter(str(?node),"%s") as ?ConnectivityNode).
        bind(strafter(str(?t),"%s") as ?Terminal).
        bind(strafter(str(?eq),"%s") as ?eq_id).

        bind(concat("{\\"@id\\":\\"", str(?eq_id),"\\",\\"@type\\":\\"",strafter(str(?eq_cls),"%s"), "\\"}") as ?Equipment)
        }
        ''' % (split, split, split, connection_params.namespace)
    # get Equipment objects associated with Container
    query_message += '''
        UNION
        {
        ?eq cim:Equipment.EquipmentContainer ?c.
        OPTIONAL {
            ?t cim:Terminal.ConductingEquipment ?eq.
            ?t cim:Terminal.ConnectivityNode ?node.
            }
        ?eq a ?eq_cls.

        bind(strafter(str(?node),"%s") as ?ConnectivityNode).
        bind(strafter(str(?t),"%s") as ?Terminal).
        bind(strafter(str(?eq),"%s") as ?eq_id).

        bind(concat("{\\"@id\\":\\"", str(?eq_id),"\\",\\"@type\\":\\"",strafter(str(?eq_cls),"%s"), "\\"}") as ?Equipment)
        }
        }
        ORDER by ?ConnectivityNode
        ''' % (split, split, split, connection_params.namespace)

    return query_message


def get_all_nodes_from_list(mrid_list: list[str], namespace: str) -> str:
    """
    Generates SPARQL query string for all nodes, terminals, and conducting equipment
    Args:

    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """
    query_message = f"""
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <{namespace}>"""
    query_message += """
        SELECT DISTINCT ?ConnectivityNode ?Terminal ?Equipment
        WHERE {
          ?c r:type cim:ConnectivityNode"""
    query_message += """
        VALUES ?ConnectivityNode {"""
    # add all equipment mRID
    for mrid in mrid_list:
        query_message += f' "{mrid}" \n'
    query_message += '               }'

    # add all attributes
    query_message += """
                {
        ?node cim:IdentifiedObject.mRID ?ConnectivityNode.

        ?t cim:Terminal.ConnectivityNode ?node.
        ?t cim:IdentifiedObject.mRID ?Terminal.
        ?t cim:Terminal.ConductingEquipment ?eq.

        ?eq cim:IdentifiedObject.mRID ?eq_id.
        ?eq a ?eq_cls.
        bind(concat("{\\"@id\\":\\"", str(?eq_id),"\\",\\"@type\\":\\"",strafter(str(?eq_cls),"%s"), "\\"}") as ?Equipment)
        }
        }
        """ % namespace

    return query_message


def get_all_nodes_from_area(area: object, connection_params: ConnectionParameters) -> str:
    """
    Generates SPARQL query string for all nodes, terminals, and conducting equipment
    Args:
        area: object of type SubSchedulingArea
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """
    # if 'SubSchedulingArea' not in connection_params.cim.__all__:
    #     _log.error("No SubSchedulingArea classes in profile")
    # else:
    #     if not isinstance(area, connection_params.cim.SubSchedulingArea):
    #         _log.error("Area is not a SubSchedulingArea")

    area_class = area.__class__.__name__
    try:
        container_uri = area.uri()
    except:
        container_uri = area.mRID

    if int(connection_params.iec61970_301) > 7:
        split = 'urn:uuid:'
    else:
        split = f'{connection_params.url}#'


    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <%s>""" % connection_params.namespace
    query_message += """
        SELECT DISTINCT ?ConnectivityNode ?Terminal ?Equipment ?Measurement
        WHERE {
          """ #
    query_message += """
        VALUES ?identifier {"%s"}
        bind(iri(concat("%s", ?identifier)) as ?c).
        """ % (container_uri, split)


    # get Equipment objects associated with Container
    query_message += '''

        ?eq cim:Equipment.SubSchedulingArea ?c.
        ?eq a ?eq_cls.

        ?t cim:Terminal.ConductingEquipment ?eq.
        ?t cim:Terminal.ConnectivityNode ?node.

        OPTIONAL {?meas cim:Measurement.Terminal ?t.
        ?meas a ?m_cls.
        bind(strafter(str(?meas),"%s") as ?meas_id).
        }

        bind(strafter(str(?node),"%s") as ?ConnectivityNode).
        bind(strafter(str(?t),"%s") as ?Terminal).

        bind(strafter(str(?eq),"%s") as ?eq_id).

        bind(concat("{\\"@id\\":\\"", str(?eq_id),"\\",\\"@type\\":\\"",strafter(str(?eq_cls),"%s"), "\\"}") as ?Equipment)
        bind(concat("{\\"@id\\":\\"", str(?meas_id),"\\",\\"@type\\":\\"",strafter(str(?m_cls),"%s"), "\\"}") as ?Measurement)

    }
        ORDER by ?Equipment
        ''' % (split, split, split, split,  connection_params.namespace, connection_params.namespace)

    return query_message



from cimgraph.queries.rdflib.get_all_attributes import \
    get_all_attributes_sparql
from cimgraph.queries.rdflib.get_all_edges import get_all_edges_sparql
from cimgraph.queries.rdflib.get_all_nodes import get_all_nodes_sparql



from __future__ import annotations

import importlib
import json
import logging
import math
import os
import re

from rdflib import Graph, Namespace, URIRef
from SPARQLWrapper import JSON, POST, SPARQLWrapper

import cimgraph.queries.ontotext as ontotext
import cimgraph.queries.sparql as sparql
from cimgraph.databases import (ConnectionInterface, ConnectionParameters,
                                QueryResponse)

_log = logging.getLogger(__name__)


class GraphDBConnection(ConnectionInterface):

    def __init__(self, connection_params: ConnectionParameters) -> None:
        self.cim_profile = connection_params.cim_profile
        self.cim = importlib.import_module('cimgraph.data_profile.' + self.cim_profile)
        self.namespace = connection_params.namespace
        self.iec61970_301 = connection_params.iec61970_301
        self.sparql_obj = None
        self.connection_parameters = connection_params

        try:
            self.data_profile = Graph(store='Oxigraph')
            path = os.path.dirname(self.cim.__file__)
            self.data_profile.parse(f'{path}/{self.cim_profile}.rdfs', format='xml')
            self.reverse = URIRef(
                'http://iec.ch/TC57/1999/rdf-schema-extensions-19990926#inverseRoleName')
        except:
            _log.warning('No RDFS schema found, reverting to default logic')
            self.data_profile = None

    def connect(self) -> None:
        if not self.sparql_obj:
            self.sparql_obj = SPARQLWrapper(self.connection_parameters.url)
            self.sparql_obj.setReturnFormat(JSON)

    def disconnect(self) -> None:
        self.sparql_obj = None

    def execute(self, query_message: str) -> QueryResponse:
        self.connect()
        self.sparql_obj.setQuery(query_message)
        self.sparql_obj.setMethod(POST)
        query_output = self.sparql_obj.query().convert()
        return query_output

    def create_new_graph(self, container: object) -> dict[type, dict[str, object]]:
        graph = {}
        # Get all nodes, terminal, and equipment by
        sparql_message = ontotext.get_all_nodes_ontotext(container, self.namespace)

        query_output = self.execute(sparql_message)

        for result in query_output['results']['bindings']:
            # Parse query results
            node_mrid = result['ConnectivityNode']['value']
            term_mrid = result['Terminal']['value']
            eq = json.loads(result['Equipment']['value'])
            eq_id = eq['@id']
            eq_class = eq['@type']
            # Add each object to graph
            node = self.create_object(graph, self.cim.ConnectivityNode, node_mrid)
            terminal = self.create_object(graph, self.cim.Terminal, term_mrid)
            if eq_class in self.cim.__all__:
                eq_class = eval(f'self.cim.{eq_class}')
                equipment = self.create_object(graph, eq_class, eq_id)

            else:
                _log.warning(f'object class missing from data profile:{eq_class}')
                continue
            # Link objects in graph
            if terminal not in equipment.Terminals:
                equipment.Terminals.append(terminal)
            if terminal not in node.Terminals:
                node.Terminals.append(terminal)
            setattr(terminal, 'ConnectivityNode', node)
            setattr(terminal, 'ConductingEquipment', equipment)

        return graph

    def get_edges_query(self, graph: dict[type, dict[str, object]], cim_class: type):

        eq_mrids = list(graph[cim_class].keys())[0:100]
        sparql_message = ontotext.get_all_edges_ontotext(cim_class, eq_mrids, self.namespace,
                                                         self.iec61970_301)

        return sparql_message

    def get_all_edges(self, graph: dict[type, dict[str, object]], cim_class: type) -> None:
        mrid_list = list(graph[cim_class].keys())
        num_nodes = len(mrid_list)
        for index in range(math.ceil(len(mrid_list) / 100)):
            eq_mrids = mrid_list[index * 100:(index + 1) * 100]
            #generate SPARQL message from correct loaders>sparql python script based on class name
            sparql_message = ontotext.get_all_edges_ontotext(cim_class, eq_mrids, self.namespace,
                                                             self.iec61970_301)
            #execute sparql query
            query_output = self.execute(sparql_message)
            self.edge_query_parser(query_output, graph, cim_class)

    def edge_query_parser(self, query_output: QueryResponse, graph: dict[type, dict[str, object]],
                          cim_class: type) -> None:
        for result in query_output['results']['bindings']:
            if result['attribute']['value'] != 'type':    #skip 'type' and other single attributes

                is_association = False
                is_enumeration = False
                mRID = result['mRID']['value']    #get mRID
                attr = result['attribute']['value']    #edge attribute
                attribute = result['attribute']['value'].split('.')    #split edge attribute
                value = result['value']['value']    #get edge value

                if self.namespace in value:    #check if enumeration
                    enum_text = value.split(self.namespace)[1]
                    enum_text = enum_text.split('>')[0]
                    enum_class = enum_text.split('.')[0]
                    enum_value = enum_text.split('.')[1]
                    is_enumeration = True

                if 'edge' in result:    #check if association
                    is_association = True
                    edge = json.loads(result['edge']['value'])
                    edge_mRID = edge['@id']
                    edge_class = edge['@type']
                    if edge_class in self.cim.__all__:
                        edge_class = eval(f'self.cim.{edge_class}')
                    else:
                        _log.warning(f'unknown class {edge_class}')
                        continue

                if is_association:    # if association to another CIM object

                    if attribute[
                            1] in cim_class.__dataclass_fields__:    #check if forward attribute
                        self.create_edge(graph, cim_class, mRID, attribute[1], edge_class,
                                         edge_mRID)

                    elif self.data_profile is not None:    # use data profile to look up reverse attribute
                        attr_uri = URIRef(f'{self.namespace}{attr}')
                        reverse_uri = self.data_profile.value(object=attr_uri,
                                                              predicate=self.reverse)
                        try:
                            reverse_attribute = reverse_uri.split('#')[1].split('.')[
                                1]    # split string
                            self.create_edge(graph, cim_class, mRID, reverse_attribute, edge_class,
                                             edge_mRID)
                        except:
                            _log.warning(f'{cim_class.__name__} does not have attribute {attr}')

                    else:    # fallback to use basic logic to identify
                        if attribute[
                                0] in cim_class.__dataclass_fields__:    #check if first name is the attribute
                            self.create_edge(graph, cim_class, mRID, attribute[0], edge_class,
                                             edge_mRID)

                        elif attribute[
                                0] + 's' in cim_class.__dataclass_fields__:    #check if attribute spelling is plural
                            self.create_edge(graph, cim_class, mRID, attribute[0] + 's',
                                             edge_class, edge_mRID)

                        elif attribute[
                                1] + 's' in cim_class.__dataclass_fields__:    #check if attribute spelling is plural
                            self.create_edge(graph, cim_class, mRID, attribute[1] + 's',
                                             edge_class, edge_mRID)

                        elif edge_class.__name__ in cim_class.__dataclass_fields__:    #check if attribute spelling is plural
                            self.create_edge(graph, cim_class, mRID, edge_class.__name__,
                                             edge_class, edge_mRID)

                        elif edge_class.__name__ + 's' in cim_class.__dataclass_fields__:    #check if attribute spelling is plural
                            self.create_edge(graph, cim_class, mRID, edge_class.__name__ + 's',
                                             edge_class, edge_mRID)

                        else:    #fallback: match class type until a suitable parent edge class is found
                            parsed = False
                            for node_attr in list(cim_class.__dataclass_fields__.keys()):
                                attr_str = cim_class.__dataclass_fields__[node_attr].type
                                edge_parent = attr_str.split('[')[1].split(']')[0]
                                if edge_parent in self.cim.__all__:
                                    parent_class = eval(f'self.cim.{edge_parent}')
                                    if issubclass(edge_class, parent_class):
                                        self.create_edge(graph, cim_class, mRID, node_attr,
                                                         edge_class, edge_mRID)
                                        parsed = True
                                        break
                            if not parsed:
                                _log.warning(f'unable to find match for {attr} for {mRID}')

                elif is_enumeration:
                    if enum_class in self.cim.__all__:    # if enumeration
                        edge_enum = eval(f'self.cim.{enum_class}(enum_value)')
                        setattr(graph[cim_class][mRID], attribute[1], edge_enum)
                else:
                    setattr(graph[cim_class][mRID], attribute[1], value)

    def create_edge(self, graph: dict[type, dict[str, object]], cim_class: type, mRID: str,
                    attribute: str, edge_class: type, edge_mRID: str) -> None:
        attribute_type = cim_class.__dataclass_fields__[attribute].type
        if 'List' in attribute_type:
            obj_list = getattr(graph[cim_class][mRID], attribute)
            found = False
            for obj in obj_list:
                if obj.mRID == edge_mRID:
                    found = True
            if not found:
                edge_object = self.create_object(graph, edge_class, edge_mRID)
                obj_list.append(edge_object)
                setattr(graph[cim_class][mRID], attribute, obj_list)
        else:
            edge_object = self.create_object(graph, edge_class, edge_mRID)
            setattr(graph[cim_class][mRID], attribute, edge_object)

    def create_object(self, graph: dict[type, dict[str, object]], class_type: type,
                      mRID: str) -> object:

        if class_type not in graph.keys():
            graph[class_type] = {}

        if mRID in graph[class_type].keys():
            obj = graph[class_type][mRID]
        else:
            obj = class_type()
            setattr(obj, 'mRID', mRID)
            graph[class_type][mRID] = obj

        return obj

    def upload(self, graph: dict[type, dict[str, object]]) -> None:
        for cim_class in graph.keys():
            for obj in graph[cim_class].values():
                query = sparql.upload_triples_sparql(obj, self.connection_params)
                self.execute(query)



import importlib
import logging
from dataclasses import dataclass, field
from uuid import UUID

from cimgraph.models.graph_model import GraphModel

_log = logging.getLogger(__name__)


@dataclass
class DistributedArea(GraphModel):

    def __post_init__(self):
        self.cim_profile = self.connection.cim_profile
        self.cim = self.connection.cim
        self.graph = {}
        self.add_to_graph(self.container)
        self.distributed_areas = []
        self.boundaries = []

    def build_from_area(self):
        self.get_all_edges(self.container.__class__)
        self.connection.create_distributed_graph(area=self.container, graph=self.graph)



    # def build_from_topo_message(self, topology_dict: dict):




#         AddrEquip = AddressableEquipment(self.cim)
#         UnaddrEquip = UnaddressableEquipment(self.cim)
#         if centralized_graph:
#             for node_mrid in topology_dict['connectivity_node']:
#                 try:
#                     uuid = UUID(node_mrid.strip('_').lower())
#                     node = centralized_graph[self.cim.ConnectivityNode][uuid]
#                     self.add_to_graph(node)
#                     for terminal in node.Terminals:
#                         self.add_to_graph(terminal)
#                         self.add_to_graph(terminal.ConductingEquipment)
#                 except:
#                     _log.warning('node ' + str(uuid) + ' not in feeder')


#         else:
#             node_list = topology_dict['connectivity_node']
#             self.connection.build_graph_from_list(self.graph, node_list)
#             self.addressable_equipment = {}
#             self.unaddressable_equipment = {}
#             AddrEquip.identify_addressable(addressable_equipment=self.addressable_equipment, graph=self.graph)
#             UnaddrEquip.identify_unaddressable(unaddressable_equipment=self.unaddressable_equipment, graph=self.graph)

#             if 'switch_areas' in topology_dict:
#                 sw_counter = -1
#                 for switch_topo in topology_dict['switch_areas']:
#                     sw_counter = sw_counter + 1
#                     node_list = switch_topo['connectivity_node']
#                     self.connection.build_graph_from_list(self.graph, node_list)
#                     # Create a new DistributedArea object for each switch area in message
#                     switch_area_id = str(
#                         self.container.mRID) + '.' + str(sw_counter)
#                     switch_container = self.cim.EquipmentContainer(mRID=switch_area_id)
#                     SwitchArea = DistributedArea(connection=self.connection,
#                                                  container=switch_container,
#                                                  distributed=True)
#                     SwitchArea.build_from_topo_message(switch_topo, self.graph)
#                     # Identify addressable and unaddressable
#                     SwitchArea.addressable_equipment = {}
#                     SwitchArea.unaddressable_equipment = {}
#                     UnaddrEquip.identify_unaddressable(
#                         unaddressable_equipment=self.unaddressable_equipment,
#                         graph=SwitchArea.graph)
#                     UnaddrEquip.identify_unaddressable(
#                         unaddressable_equipment=SwitchArea.
#                         unaddressable_equipment,
#                         graph=SwitchArea.graph)
#                     AddrEquip.identify_addressable(
#                         addressable_equipment=SwitchArea.addressable_equipment,
#                         graph=SwitchArea.graph)
#                     AddrEquip.identify_addressable(
#                         addressable_equipment=self.unaddressable_equipment,
#                         graph=SwitchArea.graph)
#                     # Add to FeederArea
#                     SwitchArea.distributed_areas = [
#                     ]  # Initialize secondary areas list
#                     self.distributed_areas.append(
#                         SwitchArea)  # Add new DistributedArea class to list

#                     if 'secondary_areas' in switch_topo:
#                         sa_counter = -1
#                         for secondary_topo in switch_topo['secondary_areas']:
#                             sa_counter = sa_counter + 1
#                             node_list = secondary_topo['connectivity_node']
#                             self.connection.build_graph_from_list(
#                                 self.graph, node_list)
#                             sec_area_id = str(self.container.mRID) + '.' + str(
#                                 sw_counter) + '.' + str(sa_counter)
#                             secondary_container = self.cim.EquipmentContainer(
#                                 mRID=sec_area_id)
#                             SecondaryArea = DistributedArea(
#                                 connection=self.connection,
#                                 container=secondary_container,
#                                 distributed=True)
#                             SecondaryArea.build_from_topo_message(
#                                 secondary_topo, self.graph)
#                             SwitchArea.distributed_areas.append(SecondaryArea)
#                             SecondaryArea.addressable_equipment = {}
#                             SecondaryArea.unaddressable_equipment = {}
#                             UnaddrEquip.identify_unaddressable(
#                                 unaddressable_equipment=self.
#                                 unaddressable_equipment,
#                                 graph=SecondaryArea.graph)
#                             UnaddrEquip.identify_unaddressable(
#                                 unaddressable_equipment=SwitchArea.
#                                 unaddressable_equipment,
#                                 graph=SecondaryArea.graph)
#                             UnaddrEquip.identify_unaddressable(
#                                 unaddressable_equipment=SecondaryArea.
#                                 unaddressable_equipment,
#                                 graph=SecondaryArea.graph)
#                             AddrEquip.identify_addressable(
#                                 addressable_equipment=self.
#                                 unaddressable_equipment,
#                                 graph=SecondaryArea.graph)
#                             AddrEquip.identify_addressable(
#                                 addressable_equipment=SwitchArea.
#                                 unaddressable_equipment,
#                                 graph=SecondaryArea.graph)
#                             AddrEquip.identify_addressable(
#                                 addressable_equipment=SecondaryArea.
#                                 addressable_equipment,
#                                 graph=SecondaryArea.graph)

#             if 'secondary_areas' in topology_dict:
#                 sa_counter = -1
#                 for secondary_topo in topology_dict['secondary_areas']:
#                     sa_counter = sa_counter + 1
#                     node_list = secondary_topo['connectivity_node']
#                     self.connection.build_graph_from_list(
#                         self.graph, node_list)
#                     secondary_container = self.cim.EquipmentContainer(
#                         mRID=self.container.mRID + '.' + str(sa_counter))
#                     SecondaryArea = DistributedArea(
#                         connection=self.connection,
#                         container=secondary_container,
#                         distributed=True)
#                     SecondaryArea.build_from_topo_message(
#                         secondary_topo, self.graph)
#                     SecondaryArea.addressable_equipment = {}
#                     SecondaryArea.unaddressable_equipment = {}
#                     UnaddrEquip.identify_unaddressable(
#                         unaddressable_equipment=self.unaddressable_equipment,
#                         graph=SecondaryArea.graph)
#                     UnaddrEquip.identify_unaddressable(
#                         unaddressable_equipment=SecondaryArea.
#                         unaddressable_equipment,
#                         graph=SecondaryArea.graph)
#                     AddrEquip.identify_addressable(
#                         addressable_equipment=self.unaddressable_equipment,
#                         graph=SecondaryArea.graph)
#                     AddrEquip.identify_addressable(
#                         addressable_equipment=SecondaryArea.
#                         addressable_equipment,
#                         graph=SecondaryArea.graph)
#                     self.distributed_areas.append(SecondaryArea)


# class DistributedTopology():

#     def __init__(self, connection: ConnectionInterface, centralized_graph,
#                  root_classes: list[type], upper_boundaries: list[type],
#                  lower_boundaries: list[type]):
#         self.connection = connection
#         self.cim = self.connection.cim
#         self.centralized_graph = centralized_graph
#         self.root_classes = root_classes
#         self.upper_boundaries = upper_boundaries
#         self.lower_boundaries = lower_boundaries

#     def create_distributed_areas(self):
#         distributed_areas = []
#         # Iterate through all types of switches
#         for root_class in self.root_classes:
#             distributed_areas = self.expand_area(root_class, distributed_areas)
#         # for DistArea in distributed_areas:
#         #     for lower_bound_class in self.lower_boundaries:

#         return distributed_areas

#     def expand_area(
#             self, root_class: type,
#             distributed_areas: list[DistributedArea]) -> list[DistributedArea]:
#         # Iterate through all instances of a switch class
#         if root_class in self.centralized_graph:
#             for equipment in self.centralized_graph[root_class].values():
#                 # Iterate through both terminals of each switch
#                 for terminal in equipment.Terminals:
#                     parsed = False
#                     for DistArea in distributed_areas:
#                         # Check if this terminal has been parsed in any other graphs
#                         if terminal.ConnectivityNode.mRID in DistArea.graph[
#                                 self.cim.ConnectivityNode].keys():
#                             parsed = True
#                     if not parsed:  # If not parsed, build switch areas from this switch
#                         NewArea = self.expand_node(terminal.ConnectivityNode)
#                         distributed_areas.append(NewArea)
#         return distributed_areas

#     def expand_node(self, root_node: object) -> DistributedArea:
#         """
#         Creates graph for a switch area
#             Args:
#                 container_mRID (str | Feeder object): The mRID of the feeder or feeder object
#                 graph (dict[type, dict[str, object]]): The typed catalog of CIM objects organized by
#                     class type and object mRID
#                 cim_class (type): The CIM class type (e.g. cim:ACLineSegment)
#             Returns:
#                 graph
#             none
#         """
#         container = self.cim.EquipmentContainer(mRID=new_mrid())
#         DistArea = DistributedArea(connection=self.connection,
#                                    container=container,
#                                    distributed=True)
#         DistArea.graph = {}
#         DistArea.add_to_graph(root_node)

#         parsed_nodes = set()
#         all_nodes = set(DistArea.graph[self.cim.ConnectivityNode].keys())
#         # all_nodes.append(node)

#         while parsed_nodes != all_nodes:
#             for node_mrid in list(all_nodes - parsed_nodes):
#                 node = DistArea.graph[self.cim.ConnectivityNode][node_mrid]
#                 if node.mRID not in parsed_nodes:

#                     for terminal in node.Terminals:
#                         equipment = terminal.ConductingEquipment
#                         DistArea.add_to_graph(terminal)
#                         DistArea.add_to_graph(equipment)

#                         if equipment.__class__ == self.cim.PowerTransformer:
#                             found_voltages = False
#                             voltages = []
#                             if equipment.PowerTransformerEnd is not None:
#                                 for end in equipment.PowerTransformerEnd:
#                                     voltages.append(end.ratedU)
#                                     found_voltages = True
#                                 if len(set(voltages)) == 1:
#                                     self.expand_terminal(terminal, DistArea)

#                             if equipment.TransformerTanks is not None and not found_voltages:
#                                 for tank in equipment.TransformerTanks:
#                                     if 'TransformerTankInfo' in self.cim.TransformerTank.__dataclass_fields__.keys(
#                                     ):
#                                         if tank.TransformerTankInfo is not None:
#                                             # Iterate through xfmr ends to get winding voltage
#                                             for end in tank.TransformerTankInfo.TransformerEndInfos:
#                                                 voltages = end.ratedU
#                                                 found_voltages = True
#                                             if len(set(voltages)) == 1:
#                                                 self.expand_terminal(
#                                                     terminal, DistArea)

#                                     if tank.Assets is not None and not found_voltages:
#                                         for asset in tank.Assets:
#                                             if asset.AssetInfo is not None:
#                                                 for end in asset.AssetInfo.TransformerEndInfos:
#                                                     voltages.append(end.ratedU)
#                                                     found_voltages = True
#                                                 if len(set(voltages)) == 1:
#                                                     self.expand_terminal(
#                                                         terminal, DistArea)

#                         # if terminal.TransformerEnd is not None:
#                         #     voltages = []
#                         #     for end in terminal.TransformerEnd:
#                         #         voltages.append(end.BaseVoltage.nominalVoltage)
#                         #         # If all xfmr ends have the same voltage, then it is a regulator
#                         #         if len(set(voltages)) == 1:
#                         #             self.expand_terminal(terminal, DistArea)
#                         #         else:
#                         #             continue

#                         if type(equipment) in self.lower_boundaries:
#                             lower_container = self.cim.EquipmentContainer(
#                                 mRID=new_mrid())
#                             LowerArea = DistributedArea(
#                                 container=lower_container,
#                                 connection=self.connection,
#                                 distributed=True)
#                             self.expand_terminal(terminal, LowerArea)
#                             DistArea.distributed_areas.append(LowerArea)

#                         elif type(equipment) in self.upper_boundaries:
#                             pass

#                         elif type(equipment):
#                             self.expand_terminal(terminal, DistArea)

#                     parsed_nodes.add(node.mRID)

#             all_nodes = set(DistArea.graph[self.cim.ConnectivityNode].keys())
#         return DistArea

#     def expand_equipment(self, equipment, DistArea):
#         for next_terminal in equipment.Terminals:
#             next_node = next_terminal.ConnectivityNode
#             if next_node not in DistArea.graph[self.cim.ConnectivityNode]:
#                 DistArea.add_to_graph(next_node)
#                 DistArea.add_to_graph(next_terminal)

#     def expand_terminal(self, root_terminal, DistArea):
#         equipment = root_terminal.ConductingEquipment
#         for next_terminal in equipment.Terminals:
#             if next_terminal != root_terminal:
#                 next_node = next_terminal.ConnectivityNode
#                 if next_node.mRID not in DistArea.graph[
#                         self.cim.ConnectivityNode].keys():
#                     DistArea.add_to_graph(next_node)
#                     DistArea.add_to_graph(next_terminal)


# def create_hierarchy_level(network: GraphModel, hierarchy: dict,
#                            top_level: bool) -> list[DistributedArea]:
#     for level in hierarchy:
#         try:
#             container_type = level['container']
#             lower_hierarchy = level['contains']
#         except:
#             _log.error(
#                 """distributed hierarchy must contain "container" and "contains" keys"""
#             )
#         # Determine container class
#         try:
#             container_class = eval(f"""network.cim.{container_type}""")
#         except:
#             _log.error('container is not a valid CIM class')

#         network.get_all_edges(container_class)
#         if container_class in network.graph:
#             containers = list(network.graph[container_class].keys())
#             for container_id in containers:
#                 container = network.graph[container_class][container_id]
#                 if not top_level:
#                     _log.info('creating new area for ' + container_type + ' ' +
#                               container.name)
#                     NewArea = DistributedArea(container=container,
#                                               connection=network.connection,
#                                               distributed=True)
#                     NewArea.distributed_areas = {}
#                     if container_class not in network.distributed_areas:
#                         network.distributed_areas[container_class] = {}
#                     network.distributed_areas[container_class][
#                         container.mRID] = NewArea
#                     NewArea.add_to_graph(container)
#                 else:
#                     NewArea = network

#                 NewArea.get_all_edges(container_class)
#                 if lower_hierarchy is not None and lower_hierarchy != []:
#                     NewArea.distributed_areas = create_hierarchy_level(
#                         NewArea, lower_hierarchy, top_level=False)
#     return network.distributed_areas


# @dataclass
# class AddressableEquipment:

#     def __init__(self, cim):
#         self.cim = cim
#         self.addressable_classes = [
#             self.cim.AsynchronousMachine, self.cim.Breaker,
#             self.cim.Disconnector, self.cim.EnergyConsumer,
#             self.cim.LinearShuntCompensator, self.cim.LoadBreakSwitch,
#             self.cim.PowerElectronicsConnection,
#             self.cim.RatioTapChanger, self.cim.Recloser,
#             self.cim.Sectionaliser, self.cim.ShuntCompensator, self.cim.Switch,
#             self.cim.SynchronousMachine
#         ]

#     def identify_addressable(self, addressable_equipment: dict,
#                              graph: dict) -> None:
#         for class_type in self.addressable_classes:
#             if class_type in graph:

#                 for uuid in graph[class_type]:
#                     addressable_equipment.append(graph[class_type][uuid])


# @dataclass
# class UnaddressableEquipment:

#     def __init__(self, cim):
#         self.cim = cim
#         self.unaddressable_classes = [
#             self.cim.ACLineSegment, self.cim.EnergySource, self.cim.Fuse,
#             self.cim.PowerTransformer, self.cim.TransformerTank
#         ]

#     def identify_unaddressable(self, unaddressable_equipment: dict,
#                                graph: dict) -> None:
#         for class_type in self.unaddressable_classes:
#             if class_type in graph:
#                 # if class_type not in unaddressable_equipment:
#                 #     unaddressable_equipment[class_type] = {}
#                 for uuid in graph[class_type]:
#                     unaddressable_equipment.append(graph[class_type][uuid])


# def create_subgeographical_area(
#         connection: ConnectionInterface,
#         subgeographicalregion: object,
#         distributed: bool = True,
#         aggregate_lower: bool = True) -> DistributedArea:
#     cim_profile = connection.connection_params.cim_profile
#     cim = importlib.import_module(f'cimgraph.data_profile.{cim_profile}')
#     SubGeographicalArea = DistributedArea(connection=connection,
#                                           container=subgeographicalregion,
#                                           distributed=distributed)
#     SubGeographicalArea.get_all_edges(cim.SubGeographicalRegion)
#     # SubGeographicalArea.get_all_edges(cim.ConnectivityNode)
#     # SubstationArea.connection.create_new_graph(substation)

#     if distributed:
#         SubGeographicalArea.distributed_areas = {}
#         # Create nested Substation areas
#         if cim.Substation in SubGeographicalArea.graph:
#             SubGeographicalArea.distributed_areas[cim.Substation] = {}
#             for substation in SubGeographicalArea.graph[
#                     cim.Substation].values():
#                 SubstationArea = create_substation_area(connection, substation)
#                 SubGeographicalArea.distributed_areas[cim.Substation][
#                     substation.mRID] = SubstationArea
#                 if aggregate_lower:
#                     aggregate_equipment(SubGeographicalArea.graph,
#                                         SubstationArea.graph)

#     return SubGeographicalArea


# def create_substation_area(connection: ConnectionInterface,
#                            substation: object,
#                            distributed=True,
#                            aggregate_lower=True) -> DistributedArea:
#     cim_profile = connection.connection_params.cim_profile
#     cim = importlib.import_module(f'cimgraph.data_profile.{cim_profile}')
#     SubstationArea = DistributedArea(connection=connection,
#                                      container=substation,
#                                      distributed=distributed)
#     SubstationArea.get_all_edges(cim.Substation)
#     SubstationArea.get_all_edges(cim.ConnectivityNode)
#     SubstationArea.connection.create_new_graph(substation)

#     if distributed:
#         # Create nested VoltageLevel areas
#         if cim.VoltageLevel in SubstationArea.graph:
#             for voltage_level in SubstationArea.graph[
#                     cim.VoltageLevel].values():
#                 VoltageLevelArea = create_voltage_level_area(
#                     connection, voltage_level)
#                 SubstationArea.distributed_areas.append(VoltageLevelArea)
#                 if aggregate_lower:
#                     aggregate_equipment(SubstationArea.graph,
#                                         VoltageLevelArea.graph)
#         # Create nested Bay areas
#         if cim.Bay in SubstationArea.graph:
#             for bay in SubstationArea.graph[cim.Bay].values():
#                 BayArea = create_bay_area(connection, bay)
#                 SubstationArea.distributed_areas.append(BayArea)
#                 if aggregate_lower:
#                     aggregate_equipment(SubstationArea.graph, BayArea.graph)

#         # Create nested Feeder areas
#         if cim.Feeder in SubstationArea.graph:
#             for feeder in SubstationArea.graph[cim.Feeder].values():
#                 FeederArea = create_feeder_area(connection, feeder)
#                 SubstationArea.distributed_areas.append(FeederArea)

#     return SubstationArea


# def create_voltage_level_area(connection: ConnectionInterface,
#                               voltage_level: object,
#                               distributed=True,
#                               aggregate_lower=True) -> DistributedArea:
#     cim_profile = connection.connection_params.cim_profile
#     cim = importlib.import_module(f'cimgraph.data_profile.{cim_profile}')
#     VoltageLevelArea = DistributedArea(connection=connection,
#                                        container=voltage_level,
#                                        distributed=distributed)
#     VoltageLevelArea.get_all_edges(cim.VoltageLevel)
#     # VoltageLevelArea.get_all_edges(cim.ConnectivityNode)
#     # VoltageLevelArea.connection.create_new_graph(voltage_level)

#     if distributed:
#         if cim.Bay in VoltageLevelArea.graph:
#             for bay in VoltageLevelArea.graph[cim.Bay].values():
#                 BayArea = DistributedArea(connection=connection,
#                                           container=bay,
#                                           distributed=True)
#                 VoltageLevelArea.distributed_areas.append(BayArea)
#                 if aggregate_lower:
#                     aggregate_equipment(VoltageLevelArea.graph, BayArea.graph)

#     return VoltageLevelArea


# def create_bay_area(connection: ConnectionInterface,
#                     bay: object,
#                     distributed=True) -> DistributedArea:
#     cim_profile = connection.connection_params.cim_profile
#     cim = importlib.import_module(f'cimgraph.data_profile.{cim_profile}')
#     BayArea = DistributedArea(connection=connection,
#                               container=bay,
#                               distributed=distributed)
#     BayArea.get_all_edges(cim.Bay)
#     BayArea.get_all_edges(cim.ConnectivityNode)
#     # VoltageLevelArea.connection.create_new_graph(bay)
#     return BayArea


# def create_feeder_area(connection: ConnectionInterface,
#                        feeder: object,
#                        distributed=True) -> DistributedArea:
#     cim_profile = connection.connection_params.cim_profile
#     cim = importlib.import_module(f'cimgraph.data_profile.{cim_profile}')
#     FeederArea = DistributedArea(connection=connection,
#                                  container=feeder,
#                                  distributed=distributed)
#     FeederArea.get_all_edges(cim.Feeder)
#     FeederArea.get_all_edges(cim.ConnectivityNode)
#     # FeederArea.connection.create_new_graph(feeder)

#     # if distributed:
#     #     for switch in FeederArea.graph[cim.LoadBreakSwitch]:
#     #         create_switch_area()
#     return FeederArea


# def aggregate_equipment(parent_graph: DistributedArea,
#                         child_graph: DistributedArea) -> None:
#     for cim_class in child_graph:
#         if cim_class not in parent_graph:
#             parent_graph[cim_class] = {}
#         parent_graph[cim_class].update(child_graph[cim_class])

#         # for obj in child_area.graph[cim_class].values():
#         #     parent_area.add_to_graph(obj)


# def create_switch_area():
#     pass


# def create_secondary_area():
#     pass



from __future__ import annotations

import concurrent.futures
import enum
import importlib
import logging
import os
from concurrent.futures import ThreadPoolExecutor
from uuid import UUID

from defusedxml.ElementTree import parse

import cimgraph.utils as cimUtils
from cimgraph.data_profile.known_problem_classes import ClassesWithManytoMany
from cimgraph.databases import (ConnectionInterface, ConnectionParameters,
                                Graph, QueryResponse)

# from cimgraph.utils.timing import timing as time_func

_log = logging.getLogger(__name__)

class XMLFile(ConnectionInterface):

    def __init__(self, connection_params: ConnectionParameters):
        self.cim_profile = connection_params.cim_profile
        self.cim = importlib.import_module('cimgraph.data_profile.' + self.cim_profile)
        self.namespace = connection_params.namespace
        self.iec61970_301 = connection_params.iec61970_301
        self.filename = connection_params.filename
        self.connection_params = connection_params
        self.graph = None
        self.connect()

    def connect(self):
        # if not graph:
        if self.filename is not None:
            try:
                self.rdf = '''{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'''
                self.namespaces = {'cim': self.namespace, 'rdf': self.rdf}
                self.tree = parse(self.filename)
                self.root = self.tree.getroot()
                self.class_index = {}
                self.graph = {}
            except:
                _log.warning(f'File {self.filename} not found. Defaulting to empty network graph')
                self.filename = None

    def disconnect(self):
        del self.tree
        del self.root
        del self.graph

    def execute(self, query_message: str) -> QueryResponse:
        pass

    def get_object(self, mrid:str, graph = {}) -> object:
        for element in self.root:
            if mrid in element.get(f'{self.rdf}about'):
                obj = self.parse_nodes(element)
        return obj

    def get_from_triple(self, subject:object, predicate:str, graph: Graph = {}) -> list[object]:
        results = []
        class_type = subject.__class__
        elements = self.tree.findall(f'.//cim:{class_type.__name__}', self.namespaces)
        for element in elements:
            if subject.uri() in element.get(f'{self.rdf}about'):
                value = element.find(f'.//cim:{predicate}', self.namespaces)
                results.append(self.parse_value(value, class_type, subject.identifier))
        return results




    def create_distributed_graph(self, area: object, graph: dict = {}) -> Graph:
        _log.error('distributed models not supported for XML file read')


    # @time_func
    def create_new_graph(self, container: object) -> Graph:

        for element in self.root:
            self.parse_nodes(element)

        with ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:

            futures = [executor.submit(self.parse_edges, element) for element in self.root]
            results = [future.result() for future in concurrent.futures.as_completed(futures)]

        return self.graph

    # async def create_new_graph_async(self):
    #     for f in asyncio.as_completed([
    #             asyncio.to_thread(self.parse_nodes(element))
    #             for element in self.root]):
    #         await f

    #     for f in asyncio.as_completed([
    #             asyncio.to_thread(self.parse_edges(element))
    #             for element in self.root]):
    #         await f

    # @time_func
    def parse_nodes(self, element:object, graph:Graph=None):
        if not graph:
            graph = self.graph

        # Iterate over the elements and create dataclass instances
        class_name = element.tag.split('{'+self.namespace+'}')[1]

        if class_name in self.cim.__all__:
            # print(class_name)
            cim_class = eval(f'self.cim.{class_name}')
            uri = element.get(f'{self.rdf}about')
            identifier = UUID(uri.strip('_').lower())
            uri = uri.split(':')[-1]  # Extract UUID from the full URI
            self.class_index[identifier] = cim_class
            obj = self.create_object(graph, cim_class, uri)
        else:
            _log.warning(f'{class_name} not in data profile')
        return obj

    # @time_func
    def parse_edges(self, element):

        class_name = element.tag.split('{'+self.namespace+'}')[1]


        if class_name in self.cim.__all__:
            cim_class = eval(f'self.cim.{class_name}')
            uri = element.get(f'{self.rdf}about')

            identifier = UUID(uri.strip('_').lower())
            uri = uri.split(':')[-1]  # Extract UUID from the full URI
            obj = self.graph[cim_class][UUID(uri)]
            for sub_element in element:
                self.parse_value(sub_element, cim_class, identifier)

        else:
            _log.warning(f'{class_name} not in data profile')

    def parse_value(self, sub_element, cim_class, identifier):

        sub_tag = sub_element.tag.split('}')[-1]
        association = self.check_attribute(cim_class, sub_tag)
        try:
            edge_uri = sub_element.attrib[f'{self.rdf}resource'].split('uuid:')[-1]
        except:
            edge_uri = None

        if edge_uri is not None:
            if self.namespace not in edge_uri:
                # try:
                    edge_uuid = UUID(edge_uri.strip('_').lower())

                    try:
                        edge_class = self.class_index[edge_uuid]
                        value = self.create_edge(self.graph, cim_class, identifier, sub_tag, edge_class, edge_uri)
                        reverse = cim_class.__dataclass_fields__[association].metadata['inverse']
                        self.create_edge(self.graph, edge_class, edge_uuid, reverse,
                                            cim_class, self.graph[cim_class][identifier].uri())
                    except:
                        value = self.get_object(edge_uri)

                # except:
                #     _log.warning(f'unable to create object with uuid {edge_uri}')
            else:
                try:
                    enum_text = edge_uri.split(self.namespace)[1]
                    enum_text = enum_text.split('>')[0]
                    enum_class = enum_text.split('.')[0]
                    enum_value = enum_text.split('.')[1]
                    edge_enum = eval(f'self.cim.{enum_class}(enum_value)')
                    if association is not None:
                        value = setattr(self.graph[cim_class][identifier], association, edge_enum)
                except:
                    pass
        else:
            if association is not None:
                value = self.create_value(self.graph, cim_class, identifier, sub_tag, sub_element.text)
        return value
        # return graph

    def parse_node_query(self, graph: dict, query_output: dict) -> Graph:
        pass

    def get_edges_query(self, graph: Graph, cim_class: type) -> str:
        pass

    def get_all_edges(self, graph: Graph, cim_class: type) -> None:
        pass

    def get_all_attributes(self, graph: Graph, cim_class: type) -> None:
        pass

    def edge_query_parser(self, query_output: QueryResponse,
                          graph: Graph, cim_class: type, expand_graph = True) -> None:
        pass



    def upload(self, graph):
        namespace = self.namespace
        iec61970_301 = self.iec61970_301
        classes_with_many_to_many = ClassesWithManytoMany()
        many_to_many = classes_with_many_to_many.attributes
        # Handling of formatting change between different 301 standard versions
        if int(iec61970_301) > 7:
            rdf_header = 'rdf:about="urn:uuid:'
            rdf_resource = 'urn:uuid:'
        else:
            rdf_header = 'rdf:ID="'
            rdf_resource = '#'
        f = open(self.filename, 'w', encoding='utf-8')
        header = '<?xml version="1.0" encoding="utf-8"?>\n'
        header += '<!-- un-comment this line to enable validation\n'
        header += '-->\n'
        header += f'<rdf:RDF xmlns:cim="{namespace}" xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n'
        header += '<!--\n'
        header += '-->\n'
        f.write(header)
        for root_class in list(graph.keys()):
            counter = 0
            for obj in graph[root_class].values():
                cim_class = obj.__class__
                header = f'<cim:{cim_class.__name__} {rdf_header}{obj.uri()}">\n'
                f.write(header)
                parent_classes = list(cim_class.__mro__)
                parent_classes.pop(len(parent_classes) - 1)
                for parent in parent_classes:
                    for attribute in parent.__annotations__.keys():
                        # Check if attribute is in data profile
                        attribute_type = cim_class.__dataclass_fields__[attribute].type
                        rdf = f'{parent.__name__}.{attribute}'
                        if attribute == 'identifier':
                            continue
                        # Upload attributes that are many-to-one or are known problem classes
                        if 'list' not in attribute_type or rdf in many_to_many:
                            edge_class = attribute_type.split('[')[1].split(']')[0]
                            edge = getattr(obj, attribute)
                            # Check if attribute is association to a class object
                            if edge_class in self.cim.__all__:
                                if edge is not None and edge != []:
                                    if type(edge.__class__) is enum.EnumMeta:
                                        resource = f'rdf:resource="{namespace}{str(edge)}"'
                                        row = f'  <cim:{parent.__name__}.{attribute} {resource}/>\n'
                                        f.write(row)
                                    elif type(edge) is str or type(edge) is bool or type(edge) is float:
                                        row = f'  <cim:{parent.__name__}.{attribute}>{str(edge)}</cim:{parent.__name__}'
                                        row += f'.{attribute}>\n'
                                        f.write(row)
                                    elif type(edge) is list:
                                        for value in edge:
                                            #TODO: lookup how to handle multiple rows of same value
                                            if type(value.__class__) is enum.EnumMeta:
                                                resource = f'rdf:resource="{namespace}{str(edge)}"'
                                                row = f'  <cim:{parent.__name__}.{attribute} {resource}/>\n'
                                                f.write(row)
                                            elif type(value) is str or type(value) is bool or type(value) is float:
                                                row = f'  <cim:{parent.__name__}.{attribute}>{str(value)}</cim:'
                                                row += f'{parent.__name__}.{attribute}>\n'
                                                f.write(row)
                                            else:
                                                resource = f'rdf:resource="{rdf_resource}{value.uri()}"'
                                                row = f'  <cim:{parent.__name__}.{attribute} {resource}/>\n'
                                                f.write(row)
                                    else:
                                        # try:
                                            resource = f'rdf:resource="{rdf_resource}{edge.uri()}"'
                                            row = f'  <cim:{parent.__name__}.{attribute} {resource}/>\n'
                                            f.write(row)
                                        # except:
                                        #     _log.warning(obj.__dict__)
                            else:
                                if edge is not None and edge != [] and rdf != 'Identity.identifier':
                                    row = f'  <cim:{parent.__name__}.{attribute}>{str(edge)}</cim:{parent.__name__}.'
                                    row += f'{attribute}>\n'
                                    f.write(row)
                tail = f'</cim:{cim_class.__name__}>\n'
                f.write(tail)
                counter = counter + 1
            _log.info(f'wrote {counter} {cim_class.__name__} objects')
        f.write('</rdf:RDF>')
        f.close()



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, is_dataclass
from uuid import UUID

from rdflib import URIRef

_log = logging.getLogger(__name__)

Graph = dict[type, dict[UUID, object]]

@dataclass
class ConnectionParameters:
    cim_profile: str = field(default_factory=str)
    url: str = field(default_factory=str)
    host: str = field(default_factory=str)
    port: str = field(default_factory=str)
    username: str = field(default_factory=str)
    password: str = field(default_factory=str)
    database: str = field(default_factory=str)
    namespace: str = field(default='http://iec.ch/TC57/CIM100#')
    iec61970_301: int = field(default=7)
    filename: str = field(default_factory=str)
    use_units: bool = field(default=False)

@dataclass
class QueryResponse:
    response: str


class ConnectionInterface(ABC):
    connection_params: ConnectionParameters

    @abstractmethod
    def connect(self):
        raise RuntimeError('Must have implemented connect in inherited class')

    @abstractmethod
    def disconnect(self):
        raise RuntimeError('Must have implemented disconnect in inherited class')

    @abstractmethod
    def execute(self, query: str) -> QueryResponse:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def upload(self, graph: Graph) -> None:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def get_all_edges(self, graph: Graph, cim_class: type) -> None:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def get_all_attributes(self, graph: Graph, cim_class: type) -> None:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def get_object(self, mRID:str, graph: Graph) -> None:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def get_from_triple(self, subject:object, predicate:str, graph: Graph = {}) -> list[object]:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def create_new_graph(self, container: object, graph: dict = {}) -> Graph:
        raise RuntimeError('Must have implemented query in the inherited class')

    @abstractmethod
    def create_distributed_graph(self, area: object, graph: dict = {}) -> Graph:
        raise RuntimeError('Must have implemented query in the inherited class')



    def check_attribute(self, cim_class:type, attribute:str) -> str:
        attr_class = attribute.split('.')[0]
        attr_link = attribute.split('.')[1]
        association = None

        if attr_link in cim_class.__dataclass_fields__:
            association = attr_link
        else:
            from_class = eval(f'self.cim.{attr_class}')
            if attr_link not in from_class.__dataclass_fields__:
                _log.warning(f'Association {attr_link} missing from class {attr_class} in data profile')

            else:
                try:
                    reverse = from_class.__dataclass_fields__[attr_link].metadata['inverse']
                    reverse_assc = reverse.split('.')[1]
                    if reverse_assc in cim_class.__dataclass_fields__:
                        association = reverse_assc
                    else:
                        _log.warning(f'Association {reverse_assc} missing from class {cim_class.__name__} in data profile')
                except:
                    _log.warning(f'Unable to find inverse of {attribute} for {cim_class.__name__}')


        return association


    def create_value(self, graph: dict[type, dict[str, object]],
                    cim_class: type, identifier: UUID, attribute: str,
                    value: str) -> bool|int|float|str:

        association = self.check_attribute(cim_class, attribute)
        if association is not None:
            attribute_type = cim_class.__dataclass_fields__[association].type
            if 'List' in attribute_type or 'list' in attribute_type:
                obj_list = getattr(graph[cim_class][identifier], association)
                if value not in str(obj_list):
                    obj_list.append(value)

            elif 'bool' in attribute_type:
                if str(value).lower() == 'true' or str(value).lower() == '1':
                    value = True
                elif str(value).lower() == 'false' or str(value).lower() == '0':
                    value = False
                else:
                    _log.warning(f'{value} for {cim_class.__name__}.{association} is not a boolean')

                setattr(graph[cim_class][identifier], association, value)


            elif 'int' in attribute_type:
                try:
                    value = int(float(value))
                except:
                    _log.warning(f'{value} for {cim_class.__name__}.{association} is not an integer')

                setattr(graph[cim_class][identifier], association, value)

            elif 'float' in attribute_type:
                try:
                    value = float(value)
                except:
                    _log.warning(f'{value} for {cim_class.__name__}.{association} is not a float')
                setattr(graph[cim_class][identifier], association, value)

            else:
                if self.connection_params.use_units:
                    pass
                    #TODO: Implement evaluation of units
                else:
                    setattr(graph[cim_class][identifier], association, value)
        return value




    def create_edge(self, graph: dict[type, dict[str, object]],
                    cim_class: type, identifier: UUID, attribute: str,
                    edge_class: type, edge_mRID: str) -> object:

        edge_object = None
        association = self.check_attribute(cim_class, attribute)
        if association is not None:
            if association in cim_class.__dataclass_fields__:
                attribute_type = cim_class.__dataclass_fields__[association].type
                if 'List' in attribute_type or 'list' in attribute_type:
                    obj_list = getattr(graph[cim_class][identifier], association)
                    if type(obj_list) is not list:
                        obj_list = [obj_list]
                    edge_uuid = (edge_mRID.strip('_').lower())
                    # _log.warning(obj_list)
                    if edge_uuid not in str(obj_list):
                        edge_object = self.create_object(graph, edge_class, edge_mRID)
                        obj_list.append(edge_object)
                        setattr(graph[cim_class][identifier], association, obj_list)
                else:
                    # _log.warning(f'{identifier}, {attribute}, {edge_class}, {edge_mRID}, ')

                    edge_object = self.create_object(graph, edge_class, edge_mRID)
                    # _log.warning(edge_object)
                    setattr(graph[cim_class][identifier], association, edge_object)
            else:
                _log.warning(f'{cim_class.__name__} does not have attribute {association}')
        return edge_object

    def create_object(self, graph:Graph,
                      class_type:type, uri:str) -> object:
        """
        Method for creating new objects and adding them to the graph
        Required Args:
            graph: an LPG graph from a GraphModel object
            class_type: a dataclass type, such as cim.ACLineSegment
            uri: the RDF ID or mRID of the object
        Returns:
            obj: a dataclass instance with the correct identifier
        """
        # Convert uri string to a uuid
        try:
            identifier = UUID(uri.strip('_').lower())
        except:
            _log.warning(f'URI {uri} for object {class_type.__name__} is not a valid UUID')
            identifier = uri

        # Add class type to graph keys if not there
        if class_type not in graph:
            graph[class_type] = {}
            # _log.warning(graph[class_type])

        # Check if object exists in graph
        if identifier in graph[class_type]:
            obj = graph[class_type][identifier]

        # If not there, create a new object and add to graph
        else:
            obj = class_type()
            obj.uuid(uri = uri)
            graph[class_type][identifier] = obj



        return obj

    def add_to_graph(self, obj: object, graph: Graph) -> None:
        """
        Method for adding existing objects to the graph
        Required Args:
            obj: a dataclass instance inheriting from the Identity class
        Returns:
            none
        """
        # Add class type to graph if not there
        if type(obj) not in graph:
            graph[type(obj)] = {}

        # Add instance to graph keys if not there
        if obj.identifier not in graph[type(obj)]:
            graph[type(obj)][obj.identifier] = obj


from cimgraph.databases.blazegraph import BlazegraphConnection
from cimgraph.databases.fileparsers import XMLFile
from cimgraph.databases.graphdb import GraphDBConnection
from cimgraph.databases.gridappsd import GridappsdConnection
from cimgraph.databases.neo4j import Neo4jConnection
from cimgraph.databases.rdflib import RDFlibConnection



from __future__ import annotations

from uuid import UUID

from cimgraph.databases import ConnectionParameters


def get_object_cypher(mrid: str, connection_params: ConnectionParameters) -> str:
    """
    Generates cypher query string to find the type of an object from its uri
    Args:
        mrid (str): The mRID or uri of the  object
        connection_params (ConnectionParameters): Database connection parameters
    Returns:
        query_message: query string that can be used in Neo4J
    """


    if int(connection_params.iec61970_301) > 7:
        split = 'urn:uuid:'
    else:
        split = f'{connection_params.url}#'

    query_message = f'''
MATCH(n)
WHERE n.uri = {split+mrid}
RETURN DISTINCT n.uri as identifier,
labels(n)[1] as class'''
    # '{"@id":"' + COALESCE(m.uri ,"") + '","@type":"' + COALESCE((labels(m))[1],"") + '"}' as edge
    return query_message



from cimgraph.models.bus_branch_model import BusBranchModel
from cimgraph.models.distributed_area import DistributedArea
from cimgraph.models.feeder_model import FeederModel
from cimgraph.models.graph_model import GraphModel
from cimgraph.models.node_breaker_model import NodeBreakerModel




from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from importlib import resources as impresources
from typing import Optional

from pint import Quantity, UnitRegistry

import cimgraph.data_profile.units.cim_units as cim_units

units_file = impresources.files(cim_units) / 'units.txt'
ureg = UnitRegistry()
ureg.load_definitions(units_file)
ureg.default_format = '~P'

def pint(obj):
    if obj.multiplier.value != 'none':
        value_str = str(obj.value) + obj.multiplier.value + obj.unit.value
    else:
        value_str = str(obj.value) + obj.unit.value
    obj.pvalue = ureg.Quantity(value_str)
    return obj.pvalue

class UnitMultiplier( Enum ):
    '''
    The unit multipliers defined for the CIM. When applied to unit symbols,
    the unit symbol is treated as a derived unit. Regardless of the contents
    of the unit symbol text, the unit symbol shall be treated as if it were
    a single-character unit symbol. Unit symbols should not contain multipliers,
    and it should be left to the multiplier to define the multiple for an entire
    data type.
    For example, if a unit symbol is 'm2Pers' and the multiplier is 'k', then
    the value is k(m**2/s), and the multiplier applies to the entire final
    value, not to any individual part of the value. This can be conceptualized
    by substituting a derived unit symbol for the unit type. If one imagines
    that the symbol 'Þ' represents the derived unit 'm2Pers', then applying
    the multiplier 'k' can be conceptualized simply as 'kÞ'.
    For example, the SI unit for mass is 'kg' and not 'g'. If the unit symbol
    is defined as 'kg', then the multiplier is applied to 'kg' as a whole and
    does not replace the 'k' in front of the 'g'. In this case, the multiplier
    of 'm' would be used with the unit symbol of 'kg' to represent one gram.
    As a text string, this violates the instructions in IEC 80000-1. However,
    because the unit symbol in CIM is treated as a derived unit instead of
    as an SI unit, it makes more sense to conceptualize the 'kg' as if it were
    replaced by one of the proposed replacements for the SI mass symbol. If
    one imagines that the 'kg' were replaced by a symbol 'Þ', then it is easier
    to conceptualize the multiplier 'm' as creating the proper unit 'mÞ', and
    not the forbidden unit 'mkg'.
    '''
    E = 'E'
    '''
    Exa 10**18.
    '''
    G = 'G'
    '''
    Giga 10**9.
    '''
    M = 'M'
    '''
    Mega 10**6.
    '''
    P = 'P'
    '''
    Peta 10**15.
    '''
    T = 'T'
    '''
    Tera 10**12.
    '''
    Y = 'Y'
    '''
    Yotta 10**24.
    '''
    Z = 'Z'
    '''
    Zetta 10**21.
    '''
    a = 'a'
    '''
    Atto 10**-18.
    '''
    c = 'c'
    '''
    Centi 10**-2.
    '''
    d = 'd'
    '''
    Deci 10**-1.
    '''
    da = 'da'
    '''
    Deca 10**1.
    '''
    f = 'f'
    '''
    Femto 10**-15.
    '''
    h = 'h'
    '''
    Hecto 10**2.
    '''
    k = 'k'
    '''
    Kilo 10**3.
    '''
    m = 'm'
    '''
    Milli 10**-3.
    '''
    micro = 'micro'
    '''
    Micro 10**-6.
    '''
    n = 'n'
    '''
    Nano 10**-9.
    '''
    none = 'none'
    '''
    No multiplier or equivalently multiply by 1.
    '''
    p = 'p'
    '''
    Pico 10**-12.
    '''
    y = 'y'
    '''
    Yocto 10**-24.
    '''
    z = 'z'
    '''
    Zepto 10**-21.
    '''

class UnitSymbol( Enum ):
    '''
    The derived units defined for usage in the CIM. In some cases, the derived
    unit is equal to an SI unit. Whenever possible, the standard derived symbol
    is used instead of the formula for the derived unit. For example, the unit
    symbol Farad is defined as 'F' instead of 'CPerV'. In cases where a standard
    symbol does not exist for a derived unit, the formula for the unit is used
    as the unit symbol. For example, density does not have a standard symbol
    and so it is represented as 'kgPerm3'. With the exception of the 'kg',
    which is an SI unit, the unit symbols do not contain multipliers and therefore
    represent the base derived unit to which a multiplier can be applied as
    a whole.
    Every unit symbol is treated as an unparseable text as if it were a single-letter
    symbol. The meaning of each unit symbol is defined by the accompanying
    descriptive text and not by the text contents of the unit symbol.
    To allow the widest possible range of serializations without requiring
    special character handling, several substitutions are made which deviate
    from the format described in IEC 80000-1. The division symbol '/' is replaced
    by the letters 'Per'. Exponents are written in plain text after the unit
    as 'm3' instead of being formatted as 'm' with a superscript of 3 or introducing
    a symbol as in 'm^3'. The degree symbol '°' is replaced with the letters
    'deg'. Any clarification of the meaning for a substitution is included
    in the description for the unit symbol.
    Non-SI units are included in list of unit symbols to allow sources of data
    to be correctly labelled with their non-SI units (for example, a GPS sensor
    that is reporting numbers that represent feet instead of meters). This
    allows software to use the unit symbol information correctly convert and
    scale the raw data of those sources into SI-based units.
    The integer values are used for harmonization with IEC 61850.
    '''
    A = 'A'
    '''
    Current in amperes.
    '''
    A2 = 'A2'
    '''
    Amperes squared (A²).
    '''
    A2h = 'A2h'
    '''
    Ampere-squared hour, ampere-squared hour.
    '''
    A2s = 'A2s'
    '''
    Ampere squared time in square amperes (A²s).
    '''
    APerA = 'APerA'
    '''
    Current, ratio of amperages. Note: Users may need to supply a prefix such
    as ‘m’ to show rates such as ‘mA/A’.
    '''
    APerm = 'APerm'
    '''
    A/m, magnetic field strength, amperes per metre.
    '''
    Ah = 'Ah'
    '''
    Ampere-hours, ampere-hours.
    '''
    As = 'As'
    '''
    Ampere seconds (A·s).
    '''
    Bq = 'Bq'
    '''
    Radioactivity in becquerels (1/s).
    '''
    Btu = 'Btu'
    '''
    Energy, British Thermal Units.
    '''
    C = 'C'
    '''
    Electric charge in coulombs (A·s).
    '''
    CPerkg = 'CPerkg'
    '''
    Exposure (x rays), coulombs per kilogram.
    '''
    CPerm2 = 'CPerm2'
    '''
    Surface charge density, coulombs per square metre.
    '''
    CPerm3 = 'CPerm3'
    '''
    Electric charge density, coulombs per cubic metre.
    '''
    F = 'F'
    '''
    Electric capacitance in farads (C/V).
    '''
    FPerm = 'FPerm'
    '''
    Permittivity, farads per metre.
    '''
    g = 'g'
    '''
    Mass in grams
    '''
    G = 'G'
    '''
    Magnetic flux density, gausses (1 G = 10-4 T).
    '''
    Gy = 'Gy'
    '''
    Absorbed dose in grays (J/kg).
    '''
    GyPers = 'GyPers'
    '''
    Absorbed dose rate, grays per second.
    '''
    H = 'H'
    '''
    Electric inductance in henrys (Wb/A).
    '''
    HPerm = 'HPerm'
    '''
    Permeability, henrys per metre.
    '''
    Hz = 'Hz'
    '''
    Frequency in hertz (1/s).
    '''
    HzPerHz = 'HzPerHz'
    '''
    Frequency, rate of frequency change. Note: Users may need to supply a prefix
    such as ‘m’ to show rates such as ‘mHz/Hz’.
    '''
    HzPers = 'HzPers'
    '''
    Rate of change of frequency in hertz per second.
    '''
    J = 'J'
    '''
    Energy in joules (N·m = C·V = W·s).
    '''
    JPerK = 'JPerK'
    '''
    Heat capacity in joules/kelvin.
    '''
    JPerkg = 'JPerkg'
    '''
    Specific energy, Joules / kg.
    '''
    JPerkgK = 'JPerkgK'
    '''
    Specific heat capacity, specific entropy, joules per kilogram Kelvin.
    '''
    JPerm2 = 'JPerm2'
    '''
    Insulation energy density, joules per square metre or watt second per square
    metre.
    '''
    JPerm3 = 'JPerm3'
    '''
    Energy density, joules per cubic metre.
    '''
    JPermol = 'JPermol'
    '''
    Molar energy, joules per mole.
    '''
    JPermolK = 'JPermolK'
    '''
    Molar entropy, molar heat capacity, joules per mole kelvin.
    '''
    JPers = 'JPers'
    '''
    Energy rate in joules per second (J/s).
    '''
    K = 'K'
    '''
    Temperature in kelvins.
    '''
    KPers = 'KPers'
    '''
    Temperature change rate in kelvins per second.
    '''
    M = 'M'
    '''
    Length, nautical miles (1 M = 1852 m).
    '''
    Mx = 'Mx'
    '''
    Magnetic flux, maxwells (1 Mx = 10-8 Wb).
    '''
    N = 'N'
    '''
    Force in newtons (kg·m/s²).
    '''
    NPerm = 'NPerm'
    '''
    Surface tension, newton per metre.
    '''
    Nm = 'Nm'
    '''
    Moment of force, newton metres.
    '''
    Oe = 'Oe'
    '''
    Magnetic field in oersteds, (1 Oe = (103/4p) A/m).
    '''
    Pa = 'Pa'
    '''
    Pressure in pascals (N/m²). Note: the absolute or relative measurement
    of pressure is implied with this entry. See below for more explicit forms.
    '''
    PaPers = 'PaPers'
    '''
    Pressure change rate in pascals per second.
    '''
    Pas = 'Pas'
    '''
    Dynamic viscosity, pascal seconds.
    '''
    Q = 'Q'
    '''
    Quantity power, Q.
    '''
    Qh = 'Qh'
    '''
    Quantity energy, Qh.
    '''
    S = 'S'
    '''
    Conductance in siemens.
    '''
    SPerm = 'SPerm'
    '''
    Conductance per length (F/m).
    '''
    Sv = 'Sv'
    '''
    Dose equivalent in sieverts (J/kg).
    '''
    T = 'T'
    '''
    Magnetic flux density in teslas (Wb/m2).
    '''
    V = 'V'
    '''
    Electric potential in volts (W/A).
    '''
    V2 = 'V2'
    '''
    Volt squared (W²/A²).
    '''
    V2h = 'V2h'
    '''
    Volt-squared hour, volt-squared-hours.
    '''
    VA = 'VA'
    '''
    Apparent power in volt amperes. See also real power and reactive power.
    '''
    VAh = 'VAh'
    '''
    Apparent energy in volt ampere hours.
    '''
    VAr = 'VAr'
    '''
    Reactive power in volt amperes reactive. The “reactive” or “imaginary”
    component of electrical power (VIsin(phi)). (See also real power and apparent
    power).
    Note: Different meter designs use different methods to arrive at their
    results. Some meters may compute reactive power as an arithmetic value,
    while others compute the value vectorially. The data consumer should determine
    the method in use and the suitability of the measurement for the intended
    purpose.
    '''
    VArh = 'VArh'
    '''
    Reactive energy in volt ampere reactive hours.
    '''
    VPerHz = 'VPerHz'
    '''
    Magnetic flux in volt per hertz.
    '''
    VPerV = 'VPerV'
    '''
    Voltage, ratio of voltages. Note: Users may need to supply a prefix such
    as ‘m’ to show rates such as ‘mV/V’.
    '''
    VPerVA = 'VPerVA'
    '''
    Power factor, PF, the ratio of the active power to the apparent power.
    Note: The sign convention used for power factor will differ between IEC
    meters and EEI (ANSI) meters. It is assumed that the data consumers understand
    the type of meter being used and agree on the sign convention in use at
    any given utility.
    '''
    VPerVAr = 'VPerVAr'
    '''
    Power factor, PF, the ratio of the active power to the apparent power.
    Note: The sign convention used for power factor will differ between IEC
    meters and EEI (ANSI) meters. It is assumed that the data consumers understand
    the type of meter being used and agree on the sign convention in use at
    any given utility.
    '''
    VPerm = 'VPerm'
    '''
    Electric field strength, volts per metre.
    '''
    Vh = 'Vh'
    '''
    Volt-hour, Volt hours.
    '''
    Vs = 'Vs'
    '''
    Volt seconds (Ws/A).
    '''
    W = 'W'
    '''
    Real power in watts (J/s). Electrical power may have real and reactive
    components. The real portion of electrical power (I&#178;R or VIcos(phi)),
    is expressed in Watts. See also apparent power and reactive power.
    '''
    WPerA = 'WPerA'
    '''
    Active power per current flow, watts per Ampere.
    '''
    WPerHz = 'WPerHz'
    '''
    Active power per change in frequency.
    '''
    WPerW = 'WPerW'
    '''
    Signal Strength, ratio of power. Note: Users may need to supply a prefix
    such as ‘m’ to show rates such as ‘mW/W’.
    '''
    WPerm2 = 'WPerm2'
    '''
    Heat flux density, irradiance, watts per square metre.
    '''
    WPerm2sr = 'WPerm2sr'
    '''
    Radiance, watts per square metre steradian.
    '''
    WPermK = 'WPermK'
    '''
    Thermal conductivity in watt/metres kelvin.
    '''
    WPers = 'WPers'
    '''
    Ramp rate in watts per second.
    '''
    WPersr = 'WPersr'
    '''
    Radiant intensity, watts per steradian.
    '''
    Wb = 'Wb'
    '''
    Magnetic flux in webers (V·s).
    '''
    Wh = 'Wh'
    '''
    Real energy in watt hours.
    '''
    anglemin = 'anglemin'
    '''
    Plane angle, minutes.
    '''
    anglesec = 'anglesec'
    '''
    Plane angle, seconds.
    '''
    bar = 'bar'
    '''
    Pressure in bars, (1 bar = 100 kPa).
    '''
    cd = 'cd'
    '''
    Luminous intensity in candelas.
    '''
    charPers = 'charPers'
    '''
    Data rate (baud) in characters per second.
    '''
    character = 'character'
    '''
    Number of characters.
    '''
    cosPhi = 'cosPhi'
    '''
    Power factor, dimensionless.
    Note 1: This definition of power factor only holds for balanced systems.
    See the alternative definition under code 153.
    Note 2 : Beware of differing sign conventions in use between the IEC and
    EEI. It is assumed that the data consumer understands the type of meter
    in use and the sign convention in use by the utility.
    '''
    count = 'count'
    '''
    Amount of substance, Counter value.
    '''
    d = 'd'
    '''
    Time in days, day = 24 h = 86400 s.
    '''
    dB = 'dB'
    '''
    Sound pressure level in decibels. Note: multiplier “d” is included in this
    unit symbol for compatibility with IEC 61850-7-3.
    '''
    dBm = 'dBm'
    '''
    Power level (logarithmic ratio of signal strength , Bel-mW), normalized
    to 1mW. Note: multiplier “d” is included in this unit symbol for compatibility
    with IEC 61850-7-3.
    '''
    deg = 'deg'
    '''
    Plane angle in degrees.
    '''
    degC = 'degC'
    '''
    Relative temperature in degrees Celsius.
    In the SI unit system the symbol is °C. Electric charge is measured in
    coulomb that has the unit symbol C. To distinguish degree Celsius from
    coulomb the symbol used in the UML is degC. The reason for not using °C
    is that the special character ° is difficult to manage in software.
    '''
    ft3 = 'ft3'
    '''
    Volume, cubic feet.
    '''
    gPerg = 'gPerg'
    '''
    Concentration, The ratio of the mass of a solute divided by the mass of
    the solution. Note: Users may need use a prefix such a ‘µ’ to express a
    quantity such as ‘µg/g’.
    '''
    gal = 'gal'
    '''
    Volume in gallons, US gallon (1 gal = 231 in3 = 128 fl ounce).
    '''
    h = 'h'
    '''
    Time in hours, hour = 60 min = 3600 s.
    '''
    ha = 'ha'
    '''
    Area, hectares.
    '''
    kat = 'kat'
    '''
    Catalytic activity, katal = mol / s.
    '''
    katPerm3 = 'katPerm3'
    '''
    Catalytic activity concentration, katals per cubic metre.
    '''
    kg = 'kg'
    '''
    Mass in kilograms. Note: multiplier “k” is included in this unit symbol
    for compatibility with IEC 61850-7-3.
    '''
    kgPerJ = 'kgPerJ'
    '''
    Weight per energy in kilograms per joule (kg/J). Note: multiplier “k” is
    included in this unit symbol for compatibility with IEC 61850-7-3.
    '''
    kgPerm3 = 'kgPerm3'
    '''
    Density in kilogram/cubic metres (kg/m³). Note: multiplier “k” is included
    in this unit symbol for compatibility with IEC 61850-7-3.
    '''
    kgm = 'kgm'
    '''
    Moment of mass in kilogram metres (kg·m) (first moment of mass). Note:
    multiplier “k” is included in this unit symbol for compatibility with IEC
    61850-7-3.
    '''
    kgm2 = 'kgm2'
    '''
    Moment of mass in kilogram square metres (kg·m²) (Second moment of mass,
    commonly called the moment of inertia). Note: multiplier “k” is included
    in this unit symbol for compatibility with IEC 61850-7-3.
    '''
    kn = 'kn'
    '''
    Speed, knots (1 kn = 1852/3600) m/s.
    '''
    l = 'l'
    '''
    Volume in litres, litre = dm3 = m3/1000.
    '''
    lPerh = 'lPerh'
    '''
    Volumetric flow rate, litres per hour.
    '''
    lPerl = 'lPerl'
    '''
    Concentration, The ratio of the volume of a solute divided by the volume
    of the solution. Note: Users may need use a prefix such a ‘µ’ to express
    a quantity such as ‘µL/L’.
    '''
    lPers = 'lPers'
    '''
    Volumetric flow rate in litres per second.
    '''
    lm = 'lm'
    '''
    Luminous flux in lumens (cd·sr).
    '''
    lx = 'lx'
    '''
    Illuminance in lux (lm/m²).
    '''
    m = 'm'
    '''
    Length in metres.
    '''
    m2 = 'm2'
    '''
    Area in square metres (m²).
    '''
    m2Pers = 'm2Pers'
    '''
    Viscosity in square metres / second (m²/s).
    '''
    m3 = 'm3'
    '''
    Volume in cubic metres (m³).
    '''
    m3Compensated = 'm3Compensated'
    '''
    Volume, cubic metres, with the value compensated for weather effects.
    '''
    m3Perh = 'm3Perh'
    '''
    Volumetric flow rate, cubic metres per hour.
    '''
    m3Perkg = 'm3Perkg'
    '''
    Specific volume, cubic metres per kilogram, v.
    '''
    m3Pers = 'm3Pers'
    '''
    Volumetric flow rate in cubic metres per second (m³/s).
    '''
    m3Uncompensated = 'm3Uncompensated'
    '''
    Volume, cubic metres, with the value uncompensated for weather effects.
    '''
    mPerm3 = 'mPerm3'
    '''
    Fuel efficiency in metres per cubic metres (m/m³).
    '''
    mPers = 'mPers'
    '''
    Velocity in metres per second (m/s).
    '''
    mPers2 = 'mPers2'
    '''
    Acceleration in metres per second squared (m/s²).
    '''
    min = 'min'
    '''
    Time in minutes, minute = 60 s.
    '''
    mmHg = 'mmHg'
    '''
    Pressure, millimetres of mercury (1 mmHg is approximately 133.3 Pa).
    '''
    mol = 'mol'
    '''
    Amount of substance in moles.
    '''
    molPerkg = 'molPerkg'
    '''
    Concentration, Molality, the amount of solute in moles and the amount of
    solvent in kilograms.
    '''
    molPerm3 = 'molPerm3'
    '''
    Concentration, The amount of substance concentration, (c), the amount of
    solvent in moles divided by the volume of solution in m³.
    '''
    molPermol = 'molPermol'
    '''
    Concentration, Molar fraction, the ratio of the molar amount of a solute
    divided by the molar amount of the solution.
    '''
    none = 'none'
    '''
    Dimension less quantity, e.g. count, per unit, etc.
    '''
    ohm = 'ohm'
    '''
    Electric resistance in ohms (V/A).
    '''
    ohmPerm = 'ohmPerm'
    '''
    Electric resistance per length in ohms per metre ((V/A)/m).
    '''
    ohmm = 'ohmm'
    '''
    Resistivity, ohm metres, (rho).
    '''
    onePerHz = 'onePerHz'
    '''
    Reciprocal of frequency (1/Hz).
    '''
    onePerm = 'onePerm'
    '''
    Wavenumber, reciprocal metres, (1/m).
    '''
    ppm = 'ppm'
    '''
    Concentration in parts per million.
    '''
    rad = 'rad'
    '''
    Plane angle in radians (m/m).
    '''
    radPers = 'radPers'
    '''
    Angular velocity in radians per second (rad/s).
    '''
    radPers2 = 'radPers2'
    '''
    Angular acceleration, radians per second squared.
    '''
    rev = 'rev'
    '''
    Amount of rotation, revolutions.
    '''
    rotPers = 'rotPers'
    '''
    Rotations per second (1/s). See also Hz (1/s).
    '''
    s = 's'
    '''
    Time in seconds.
    '''
    sPers = 'sPers'
    '''
    Time, Ratio of time. Note: Users may need to supply a prefix such as ‘&#181;’
    to show rates such as ‘&#181;s/s’.
    '''
    sr = 'sr'
    '''
    Solid angle in steradians (m2/m2).
    '''
    therm = 'therm'
    '''
    Energy, therms.
    '''
    tonne = 'tonne'
    '''
    Mass in tons, “tonne” or “metric ton” (1000 kg = 1 Mg).
    '''

@dataclass
class ActivePower():
    '''
    Product of RMS value of the voltage and the RMS value of the in-phase
    component of the current.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.W)
    def __post_init__(self):
        pint(self)

@dataclass
class ActivePowerChangeRate():
    '''
    Rate of change of active power per time.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.WPers)
    def __post_init__(self):
        pint(self)

@dataclass
class ActivePowerPerCurrentFlow():
    '''
    Active power variation with current flow.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.WPerA)
    def __post_init__(self):
        pint(self)

@dataclass
class ActivePowerPerFrequency():
    '''
    Active power variation with frequency.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.WPerHz)
    def __post_init__(self):
        pint(self)

@dataclass
class Admittance():
    '''
    Ratio of current to voltage.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.S)
    def __post_init__(self):
        pint(self)

@dataclass
class AngleDegrees():
    '''
    Ratio of current to voltage.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.deg
    def __post_init__(self):
        pint(self)

@dataclass
class AngleRadians():
    '''
    Ratio of current to voltage.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.rad
    def __post_init__(self):
        pint(self)

@dataclass
class ApparentPower():
    '''
    Product of the RMS value of the voltage and the RMS value of the current.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.VA)
    def __post_init__(self):
        pint(self)

@dataclass
class Area():
    '''
    Area.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.m2
    def __post_init__(self):
        pint(self)

@dataclass
class Capacitance():
    '''
    Capacitive part of reactance (imaginary part of impedance), at rated frequency.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.F)
    def __post_init__(self):
        pint(self)

@dataclass
class CapacitancePerLength():
    '''
    Capacitance per unit length.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.FPerm)
    def __post_init__(self):
        pint(self)

@dataclass
class Classification():
    '''
    Classification of level.  Specify as 1..n, with 1 being the most detailed,
    highest priority, etc as described on the attribute using this data type.
    '''
    value: int = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.rad
    def __post_init__(self):
        pint(self)

@dataclass
class Conductance():
    '''
    Factor by which voltage must be multiplied to give corresponding power lost
    from a circuit. Real part of admittance.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.S)
    def __post_init__(self):
        pint(self)

@dataclass
class ConductancePerLength():
    '''
    Real part of admittance per unit of length.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.SPerm)
    def __post_init__(self):
        pint(self)

@dataclass
class CurrentFlow():
    '''
    Electrical current with sign convention: positive flow is out of the
    conducting equipment into the connectivity node. Can be both AC and DC.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.A)
    def __post_init__(self):
        pint(self)

@dataclass
class Damping():
    '''
    Per-unit active power variation with frequency referenced on the system
    apparent power base. Typical values are in the range 1,0 - 2,0.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.onePerHz
    def __post_init__(self):
        pint(self)

@dataclass
class Date():
    '''
    Date as "yyyy-mm-dd", which conforms with ISO 8601. UTC time zone is
    specified as "yyyy-mm-ddZ". A local timezone relative UTC is specified as
    "yyyy-mm-dd(+/-)hh:mm".
    '''

@dataclass
class DateInterval():
    '''
    Interval between two dates.
    '''
    start: MonthDay = field(default=None)
    end: MonthDay = field(default=None)

@dataclass
class Displacement():
    '''
    Unit of displacement relative to a reference position,
    hence can be negative.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.m
    def __post_init__(self):
        pint(self)

@dataclass
class Duration():
    '''
    Duration as "PnYnMnDTnHnMnS" which conforms to ISO 8601, where nY expresses
    a number of years, nM a number of months, nD a number of days. The letter
    T separates the date expression from the time expression and, after it,
    nH identifies a number of hours, nM a number of minutes and nS a number
    of seconds. The number of seconds could be expressed as a decimal number,
    but all other numbers are integers.
    '''

@dataclass
class Emission():
    '''
    Quantity of emission per fuel heat content.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.kgPerJ
    def __post_init__(self):
        pint(self)

@dataclass
class Frequency():
    '''
    Cycles per second.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.Hz
    def __post_init__(self):
        pint(self)

@dataclass
class HeatRate():
    '''
    Heat generated, in energy per time unit of elapsed time.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.JPers
    def __post_init__(self):
        pint(self)

@dataclass
class Hours():
    '''
    Time specified in hours
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.h
    def __post_init__(self):
        pint(self)

@dataclass
class Impedance():
    '''
    Ratio of voltage to current.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.ohm)
    def __post_init__(self):
        pint(self)

@dataclass
class Inductance():
    '''
    Inductive part of reactance (imaginary part of impedance), at rated frequency.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.H)
    def __post_init__(self):
        pint(self)


@dataclass
class InductancePerLength():
    '''
    Inductance per unit of length.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.HPerm)
    def __post_init__(self):
        pint(self)

@dataclass
class KiloActivePower():
    '''
    Inductance per unit of length.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.k)
    unit: UnitSymbol = field(default=UnitSymbol.W)
    def __post_init__(self):
        pint(self)

@dataclass
class Length():
    '''
    Unit of length. It shall be a positive value or zero.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.m
    def __post_init__(self):
        pint(self)

@dataclass
class Mass():
    '''
    Heat generated, in energy per time unit of elapsed time.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.k)
    @property #read-only
    def unit(self):
        return UnitSymbol.g
    def __post_init__(self):
        pint(self)

@dataclass
class Minutes():
    '''
    Time in minutes.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.k)
    @property #read-only
    def unit(self):
        return UnitSymbol.min
    def __post_init__(self):
        pint(self)

@dataclass
class MonthDay():
    '''
    MonthDay format as "--mm-dd", which conforms with XSD data type gMonthDay.
    '''

@dataclass
class MonthDayInterval():
    '''
    Interval between two times specified as month and day.
    '''
    start: MonthDay = field(default=None)
    end: MonthDay = field(default=None)

@dataclass
class PU():
    '''
    Per Unit - a positive or negative value referred to a defined base. Values typically range from -10 to +10.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.none)
    def __post_init__(self):
        pint(self)

@dataclass
class PerCent():
    '''
    Percentage on a defined base. For example, specify as 100 to indicate at the defined base.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.none
    def __post_init__(self):
        pint(self)

@dataclass
class Pressure():
    '''
    Time in minutes.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.Pa
    def __post_init__(self):
        pint(self)

@dataclass
class Reactance():
    '''
    Reactance (imaginary part of impedance), at rated frequency.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.ohm)
    def __post_init__(self):
        pint(self)


@dataclass
class ReactancePerLength():
    '''
    Reactance (imaginary part of impedance) per unit of length, at rated frequency.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.ohmPerm)
    def __post_init__(self):
        pint(self)

@dataclass
class ReactivePower():
    '''
    Product of RMS value of the voltage and the RMS value of the quadrature
    component of the current.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.VAr)
    def __post_init__(self):
        pint(self)

@dataclass
class Resistance():
    '''
    Resistance (real part of impedance).
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.ohm)
    def __post_init__(self):
        pint(self)

@dataclass
class ResistancePerLength():
    '''
    Resistance (real part of impedance) per unit of length.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.ohmPerm)
    def __post_init__(self):
        pint(self)

@dataclass
class RotationSpeed():
    '''
    Number of revolutions per second.
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.none
    def __post_init__(self):
        pint(self)

@dataclass
class Seconds():
    '''
    Time specified in seconds
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.s
    def __post_init__(self):
        pint(self)

@dataclass
class Speed():
    '''
    Distance per unit of time
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.mPers
    def __post_init__(self):
        pint(self)

@dataclass
class Susceptance():
    '''
    Imaginary part of admittance.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.S)
    def __post_init__(self):
        pint(self)

@dataclass
class SusceptancePerLength():
    '''
    Imaginary part of admittance per unit of length.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.SPerm)
    def __post_init__(self):
        pint(self)

@dataclass
class Temperature():
    '''
    Distance per unit of time
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.degC
    def __post_init__(self):
        pint(self)

@dataclass
class Time():
    '''
    Time as "hh:mm:ss.sss", which conforms with ISO 8601. UTC time zone is
    specified as "hh:mm:ss.sssZ". A local timezone relative UTC is specified
    as "hh:mm:ss.sss±hh:mm". The second component (shown here as "ss.sss")
    could have any number of digits in its fractional part to allow any
    kind of precision beyond seconds.
    '''

@dataclass
class TimeInterval():
    '''
    Interval between two times.
    '''
    start: MonthDay = field(default=None)
    end: MonthDay = field(default=None)

@dataclass
class Voltage():
    '''
    Electrical voltage, can be both AC and DC.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.V)
    def __post_init__(self):
        pint(self)

@dataclass
class VoltagePerReactivePower():
    '''
    Voltage variation with reactive power.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    unit: UnitSymbol = field(default=UnitSymbol.VPerVAr)
    def __post_init__(self):
        pint(self)

@dataclass
class Volume():
    '''
    Volume.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.m3
    def __post_init__(self):
        pint(self)

@dataclass
class VolumeFlowRate():
    '''
    Volume per time
    '''
    value: float = field(default=None)
    @property #read-only
    def multiplier(self):
        return UnitMultiplier.none
    @property #read-only
    def unit(self):
        return UnitSymbol.m3Pers
    def __post_init__(self):
        pint(self)

@dataclass
class WaterLevel():
    '''
    Reservoir water level referred to a given datum such as mean sea level.
    '''
    value: float = field(default=None)
    multiplier: UnitMultiplier = field(default=UnitMultiplier.none)
    @property #read-only
    def unit(self):
        return UnitSymbol.m
    def __post_init__(self):
        pint(self)



"""
This module provides functionality to interact with the GridAPPSD platform.

It attempts to import the GridAPPSD package and sets an environment variable
'CIMGRAPH_HAS_GRIDAPPSD' to '1' and HAS_GRIDAPPSD flag set to True, if the import is
successful. If the import fails, it sets a flag 'HAS_GRIDAPPSD' to False.

The module also provides a function 'get_topology_response' which sends a
request to the GridAPPSD platform to get the topology of a specific feeder.

Functions:
----------
get_topology_response(feeder_mrid: str, gapps: GridAPPSD) -> Dict:
    Sends a request to the GridAPPSD platform to get the topology of a specific feeder.

    Parameters:
    -----------
    feeder_mrid : str
        The mRID of the feeder for which the topology is requested.
    gapps : GridAPPSD
        An instance of the GridAPPSD class.

    Returns:
    --------
    dict
        A dictionary containing the response from the GridAPPSD platform.
"""
from __future__ import annotations

from gridappsd import GridAPPSD

from cimgraph.databases.gridappsd.gridappsd import GridappsdConnection


def get_topology_response(feeder_mrid: str, gapps: GridAPPSD) -> dict:
    assert feeder_mrid is not None

    topic = 'goss.gridappsd.request.data.topology'
    message = {'requestType': 'GET_SWITCH_AREAS', 'modelID': feeder_mrid, 'resultFormat': 'JSON'}

    topo_response = gapps.get_response(topic=topic, message=message, timeout=30)
    return topo_response



import json
import logging
from dataclasses import dataclass, field, is_dataclass
from random import Random
from typing import Optional
from uuid import UUID, uuid4

# from cimgraph.utils.timing import timing as time_func

ARCHIVE_JSON_LD = True

_log = logging.getLogger(__name__)


class UUID_Meta():
    uuid:UUID = None
    uuid_str:str = ''
    uri_has_underscore:bool = False
    uri_is_capitalized:bool = False
    mrid_has_underscore:bool = False
    mrid_is_capitalized:bool = False

    # Create UUID from inconsistent mRIDs
    def generate_uuid(self, mRID:str = None, uri:str = None, name:str = None, seed:str = None) -> UUID:
        if seed is None:
            seed = ''
        invalid_mrid = True

        # If URI is specified, try creating from UUID from URI
        if uri is not None:
            # Handle inconsistent capitalization / underscores
            if uri.strip('_') != uri:
                self.uri_has_underscore = True
            if uri.lower() != uri:
                self.uri_is_capitalized = True
            try:
                self.uuid = UUID(uri.strip('_').lower())
                identifier = self.uuid
                invalid_mrid = False
            except:
                seed = seed + uri
                _log.warning(f'URI {uri} not a valid UUID, generating new UUID')
                mRID = str(uri)
            else:
                self.uuid = identifier
                invalid_mrid = False

        # If URI is specified, try creating from UUID from URI
        if mRID is not None:
            # Handle inconsistent capitalization / underscores
            if mRID.strip('_') != mRID:
                self.mrid_has_underscore = True
                if uri is None:
                    self.uri_has_underscore = True
            if mRID.lower() != mRID:
                self.mrid_is_capitalized = True
                if uri is None:
                    self.uri_is_capitalized = True
            # Create a new UUID based on the mRID if it does not exist
            if self.uuid is None:
                try:
                    identifier = UUID(mRID.strip('_').lower())
                    invalid_mrid = False
                except:
                    seed = seed + mRID
                    _log.warning(f'mRID {mRID} not a valid UUID, generating new UUID')

        # Otherwise, build UUID using unique name as a seed
        if invalid_mrid:

            if seed:
                randomGenerator = Random(seed)
                self.uuid = UUID(int=randomGenerator.getrandbits(128), version=4)
            else:
                self.uuid = uuid4()

            identifier = self.uuid

        self.uuid_str = str(identifier)
        return identifier

@dataclass
class Identity():
    '''
    This is the new root class from CIM 18 to provide common identification
    for all classes needing identification and naming attributes.
    IdentifiedObject is now a child class of Identity.
    mRID is superseded by Identity.identifier, which is typed to be a UUID.
    '''
    identifier: Optional[UUID] = field(
        default = None,
        metadata = {
            'type': 'Attribute',
            'minOccurs': '1',
            'maxOccurs': '1'
        })

    # Backwards support for objects created with mRID
    # @time_func
    def __post_init__(self) -> None:
        # Validate if pre-specified
        if self.identifier is not None:
            id = str(self.identifier)
            # Check if Identity.identifier is an invalid UUID
            if not isinstance(self.identifier,UUID):
                _log.warning(f'Identifier {self.identifier} must be a UUID object, generating new UUID')
                self.identifier = None
            # If object inherits from IdentifiedObject, create uuid from mRID
            if 'mRID' in self.__dataclass_fields__:
                self.uuid(uri=id, mRID=self.mRID, name=self.name)
            else:
                self.uuid(uri=id)
        # Otherwise, create a new UUID
        else:
            if 'mRID' in self.__dataclass_fields__:
                self.uuid(mRID=self.mRID, name=self.name)
            else:
                self.uuid()



    # Override python string for printing with JSON representation
    # @time_func
    def __str__(self, show_mRID:bool = False, show_empty:bool = False,
                use_names:bool = False) -> json:
        # Create JSON-LD dump with repr and all attributes
        dump = dict(json.loads(self.__repr__()) | self.__dict__)
        del dump['__uuid__']
        del dump['__json_ld__']
        attribute_list = list(self.__dataclass_fields__.keys())
        for attribute in attribute_list:
            # Delete attributes from print that are empty
            if dump[attribute] is None or dump[attribute] == []:
                if not show_empty:
                    del dump[attribute]
            # If a dataclass, replace with custom repr
            elif is_dataclass(dump[attribute]):
                if use_names and 'name' in dump[attribute].__dataclass_fields__:
                    dump[attribute] = dump[attribute].name
                else:
                    dump[attribute] = dump[attribute].__repr__()
            # Reformat all attributes as string for JSON
            elif not isinstance(dump[attribute], str):
                dump[attribute] = str(dump[attribute])
        # Remove duplicate identifier and mRID from JSON-LD
        if not show_mRID:
            del dump['identifier']
            if 'mRID' in dump:
                del dump['mRID']
        # Fix python ' vs JSON "
        dump = json.dumps(dump)
        dump = str(dump).replace('\\\"','\"' )
        dump = str(dump).replace('\"[','[' )
        dump = str(dump).replace(']\"',']' )
        dump = str(dump).replace('\"{','{' )
        dump = str(dump).replace('}\"','}' )
        return dump

    # Override python __repr__ method with JSON-LD representation
    # This is needed to avoid infinite loops in object previews
    # @time_func
    def __repr__(self) -> str:
        if ARCHIVE_JSON_LD:
            return self.__json_ld__
        else:
            return json.dumps({'@id': f'{str(self.identifier)}', '@type': f'{self.__class__.__name__}'})

    # Add indentation of json for pretty print
    def pprint(self, print_mRID:bool=False) -> None:
        print(json.dumps(json.loads(self.__str__(print_mRID)), indent=4))

    # Create UUID from inconsistent mRIDs
    # @time_func
    def uuid(self, mRID:str = None, uri:str = None, name:str = None, seed:str = None) -> UUID:
        self.__uuid__ = UUID_Meta()
        if seed is None:
            seed = ''
        if name is not None:
            seed = seed + f'{self.__class__.__name__}:{name}'

        self.identifier = self.__uuid__.generate_uuid(mRID=mRID, uri=uri, name=name, seed=seed)

        # Write mRID string for backwards compatibility
        if 'mRID' in self.__dataclass_fields__:
            if mRID is not None:
                self.mRID = mRID
            else:
                self.mRID = str(self.identifier)
            if name is not None:
                self.name = name

        if ARCHIVE_JSON_LD:
            self.__json_ld__ = json.dumps({'@id': f'{str(self.identifier)}', '@type': f'{self.__class__.__name__}'})


    # Method to reconstitute URI from UUID
    def uri(self) -> str:
        uri = str(self.identifier)
        try:
            if self.__uuid__.uri_is_capitalized:
                uri = uri.upper()
            if self.__uuid__.uri_has_underscore:
                uri = '_' + uri
        except:
            pass
        return uri



import logging
from dataclasses import dataclass, field

from cimgraph.models.distributed_area import DistributedArea
from cimgraph.models.graph_model import GraphModel

# from cimgraph.utils.timing import timing as time_func

_log = logging.getLogger(__name__)


@dataclass
class FeederModel(GraphModel):
    """
    Knowledge graph class for distribution feeder objects. This should class
    should be used for all feeder models and distribution data.
    Args:
        container: a CIM Feeder object with specified mRID
        connection: a ConnectionInterface object, such as BlazegraphConnection
        distributed: a boolean to indicate if the graph is distributed
    Optional Args:
        distributed_hierarchy: Custom inheritance structure for defining distributed areas
        topology_message: JSON message from GridAPPS-D Topology Proccessor Service
    Returns:
        none
    Methods:
        add_to_graph(object): adds a new CIM object to the knowledge graph
        get_all_edges(cim.ClassName): universal database query to expand graph by one edge
        graph[cim.ClassName]: access to graph dictionary sorted by class and mRID
        pprint(cim.ClassName): pretty-print method for showing graph of a class type
        get_edges_query(cim.ClassName): returns query text for debugging
    """
    topology_message: dict = field(default_factory=dict)
    distributed_hierarchy: list[type] = field(default_factory=list)
    distributed_areas: list[DistributedArea] | None = None

    def __post_init__(self):

        if self.connection is not None:    # Check if connection has been specified
            self.cim = self.connection.cim    # Set CIM data profile
            if self.distributed:    # Check if distributed flag is true
                # Build distributed network model
                self.initialize_distributed_model(self.container)
            else:
                # Otherwise build centralized network model
                self.initialize_centralized_model(self.container)
        else:    # Log error thant no connection was specified
            _log.error('A ConnectionInterface must be specified')
    # @time_func
    def initialize_centralized_model(self, container: object) -> None:
        # Build graph model using database-specific routine
        self.graph = self.connection.create_new_graph(container)

    def initialize_distributed_model(self, container: object) -> None:
        self.distributed_areas = []

        if not isinstance(container, self.cim.Feeder):
            error_message = 'Invalid argument: container must be an instance of cim.Feeder'
            _log.error(error_message)
            raise TypeError(error_message)

        self.add_to_graph(container)
        new_edges = self.get_from_triple(container, 'Feeder.FeederArea')

        if not new_edges:
            error_message = f'No FeederArea defined for Feeder {container.uri()}. '
            error_message += 'Rebuild the model with the create_distributed_feeder() method'
            error_message += 'from the CIM-Graph-Topology-Processor library.'
            _log.error(error_message)
            raise ValueError(error_message)

        feeder_area = new_edges[0]
        # Create a new DistributedArea GraphModel for the feeder area
        feeder_area_model = DistributedArea(container=feeder_area,
                                              connection=self.connection,
                                              distributed=True)

        # Initialize DistributedArea with equipment, nodes, terminals, and measurements
        feeder_area_model.get_all_edges(self.cim.FeederArea)
        feeder_area_model.build_from_area()
        # Append to distributed_areas field of FeederModel GraphModel.
        self.distributed_areas.append(feeder_area_model)

        for switch_area in feeder_area.SwitchAreas:
            # Create a new DistributedArea GraphModel for each switch area
            switch_area_model = DistributedArea(container=switch_area,
                                              connection=self.connection,
                                              distributed=True)
            # Initialize DistributedArea with equipment, nodes, terminals, and measurements
            switch_area_model.get_all_edges(self.cim.SwitchArea)
            switch_area_model.build_from_area()
            # Add switch area context object to list of dist areas for feeder area context
            feeder_area_model.distributed_areas.append(switch_area_model)

            for secondary_area in switch_area.SecondaryAreas:
                # Create new DistributedArea GraphModel for each secondary area
                secondary_area_model = DistributedArea(container=secondary_area,
                                                connection=self.connection,
                                                distributed=True)
                # Initialize DistributedArea with equipment, nodes, terminals, and measurements
                secondary_area_model.build_from_area()
                # Add secondary area context to list of dist areas for switch area context
                switch_area_model.distributed_areas.append(secondary_area_model)



from __future__ import annotations

from uuid import UUID

from cimgraph.databases import ConnectionParameters


def get_object_sparql(mrid: str, connection_params: ConnectionParameters) -> str:
    """
    Generates SPARQL query string to find the type of an object from its uri
    Args:
        mrid (str): The mRID or uri of the  object
        connection_params (ConnectionParameters): Database connection parameters
    Returns:
        query_message: query string that can be used in blazegraph connection or STOMP client
    """

    if int(connection_params.iec61970_301) > 7:
        split = 'urn:uuid:'
    else:
        split = f'{connection_params.url}#'

    query_message = """
        PREFIX r:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX cim:  <%s>""" % connection_params.namespace

    query_message += """
        SELECT DISTINCT ?identifier ?obj_class
        WHERE {
          """

    query_message += '''
    VALUES ?identifier {"%s"}''' %mrid
    # add all equipment mRID

    query_message += f'''
        bind(iri(concat("{split}", ?identifier)) as ?eq)'''

    query_message += """

        ?eq a ?classraw.
        bind(strafter(str(?classraw),"%s") as ?obj_class)
        }
        ORDER by  ?identifier
        """ % (connection_params.namespace)
    return query_message



from __future__ import annotations

import logging
from dataclasses import dataclass, field

from cimgraph.databases import ConnectionInterface
from cimgraph.models.distributed_area import DistributedArea
from cimgraph.models.graph_model import GraphModel

_log = logging.getLogger(__name__)


@dataclass
class NodeBreakerModel(GraphModel):
    """
    Knowledge graph class for transmission node-breaker models. This should class
    should be used for all models with detailed substation representations.
    Args:
        container: a CIM object inheriting from EquipmentContainer with specified mRID
        connection: a ConnectionInterface object, such as BlazegraphConnection
        distributed: a boolean to indicate if the graph is distributed
    Optional Args:
        distributed_hierarchy: Custom inheritance structure for defining distributed areas
        topology_message: JSON message from GridAPPS-D Topology Proccessor Service
    Returns:
        none
    Methods:
        graph[cim.ClassName]: access to graph dictionary sorted by class and mRID
        add_to_graph(object): adds a new CIM object to the knowledge graph
        get_all_edges(cim.ClassName): universal database query to expand graph by one edge
        get_all_attributes(cim.ClassName): universal query to get attributes without expanding graph
        get_edges_query(cim.ClassName): returns query text for debugging
        pprint(cim.ClassName): pretty-print method for showing graph of a class type

    """
    # topology_message: dict = field(default_factory=dict)
    # distributed_hierarchy: list[type] = field(default_factory=list)
    aggregate_lower_areas: bool = field(default=True)

    def __post_init__(self):
        if self.connection is not None:    # Check if connection has been specified
            self.cim = self.connection.cim    # Set CIM data profile
            if self.distributed:    # Check if distributed flag is true
                # Build distributed network model
                self.initialize_distributed_model(self.container)
            else:
                # Otherwise build centralized network model
                self.initialize_centralized_model(self.container)
        else:    # Log error thant no connection was specified
            _log.error('A ConnectionInterface must be specified')

    def initialize_centralized_model(self, container: object) -> None:
        self.graph = self.connection.create_new_graph(container)
        self.add_to_graph(container)

    def initialize_distributed_model(self, container: object) -> None:
        self.graph = {}
        self.add_to_graph(self.container)
        self.distributed_areas = {}

        # if container.__class__ == self.cim.GeographicalRegion:
        #     self.get_all_edges(self.cim.GeographicalRegion)
        #     self.distributed_areas[self.cim.SubGeographicalRegion] = {}
        #     self.distributed_areas[self.cim.Substation] = {}
        #     sub_geo_list = list(
        #         self.graph[self.cim.SubGeographicalRegion].keys())
        #     for sub_geo_mrid in sub_geo_list:
        #         sub_geo = self.graph[
        #             self.cim.SubGeographicalRegion][sub_geo_mrid]
        #         SubGeographicalArea = create_subgeographical_area(
        #             self.connection, sub_geo)
        #         self.distributed_areas[self.cim.SubGeographicalRegion][
        #             sub_geo_mrid] = SubGeographicalArea
        #         if self.aggregate_lower_areas:
        #             aggregate_equipment(self.graph, SubGeographicalArea.graph)
        #             self.distributed_areas.update(
        #                 SubGeographicalArea.distributed_areas)

        # elif container.__class__ == self.cim.SubGeographicalRegion:
        #     self.get_all_edges(self.cim.SubGeographicalRegion)

        #     for substation in self.graph[self.cim.Substation].values():
        #         SubstationArea = create_substation_area(
        #             self.connection, substation)
        #         self.distributed_areas.append(SubstationArea)
        #         if self.aggregate_lower_areas:
        #             aggregate_equipment(self.graph, SubstationArea.graph)

        # self.distristributed_areas = create_hierarchy_level(self, self.distributed_hierarchy, top_level=True)



'''
    Annotated CIMantic Graphs data profile init file for cim17v40
    Generated by CIMTool http://cimtool.org
'''
from cimgraph.data_profile.cimhub_ufls.cimhub_ufls import (
    PU, Accumulator, AccumulatorLimit, AccumulatorLimitSet, AccumulatorReset,
    AccumulatorValue, ACDCTerminal, ACLineSegment, ACLineSegmentPhase,
    ActivePower, ActivePowerLimit, ActivePowerPerFrequency, Analog,
    AnalogControl, AnalogLimit, AnalogLimitSet, AnalogValue, AngleDegrees,
    AngleRadians, ApparentPower, ApparentPowerLimit, Area, AreaConfiguration,
    Asset, AssetContainer, AssetInfo, AssetKind, AssetOrganisationRole,
    AsynchronousMachine, AsynchronousMachineKind, BaseFrequency, BasePower,
    BaseVoltage, BasicIntervalSchedule, BatteryStateKind, BatteryUnit, Bay,
    BranchGroup, BranchGroupTerminal, Breaker, BreakerConfiguration,
    BusbarConfiguration, BusbarSection, BusbarSectionInfo, Bushing,
    BushingInfo, BushingInsulationKind, BusNameMarker, CableConstructionKind,
    CableInfo, CableOuterJacketKind, CableShieldMaterialKind, Capacitance,
    CatalogAssetType, Clamp, Command, CompositeSwitch,
    ConcentricNeutralCableInfo, Conductance, ConductancePerLength,
    ConductingEquipment, Conductor, ConformLoad, ConformLoadGroup,
    ConformLoadSchedule, ConnectivityNode, ConnectivityNodeContainer,
    Connector, Control, ConverterControlModeKind, CoolantType,
    CoordinateSystem, Currency, CurrentFlow, CurrentLimit, Curve, CurveData,
    CurveStyle, Customer, CustomerAgreement, CustomerKind, Cut, DateInterval,
    DateTimeInterval, DayType, DecimalQuantity, DemandResponseProgram,
    DERCurveData, DERDynamics, DERFunction, DERGroupDispatch, DERGroupForecast,
    DERMonitorableParameter, DERParameterKind, DERUnitSymbol, Disconnector,
    Discrete, DiscreteValue, DispatchSchedule, Displacement, DistributionArea,
    DuctBank, EarthFaultCompensator, EndDevice, EndDeviceControl,
    EndDeviceGroup, EnergyArea, EnergyConnection, EnergyConnectionProfile,
    EnergyConsumer, EnergyConsumerPhase, EnergySource, EnergySourcePhase,
    Equipment, EquipmentContainer, ExternalNetworkInjection, Feeder,
    FeederArea, FloatQuantity, FlowDirectionKind, Frequency,
    FrequencyConverter, FrequencyProtectionFunctionBlock, FunctionBlock,
    FunctionInputVariable, FunctionOutputVariable, Fuse, GeneratingUnit,
    GeneratorControlMode, GeneratorControlSource, GenUnitOpSchedule,
    GeographicalRegion, GrossToNetActivePowerCurve, Ground, GroundDisconnector,
    GroundingImpedance, House, HouseCooling, HouseHeating,
    HouseThermalIntegrity, HydroPump, IdentifiedObject, Identity,
    IEC61970CIMVersion, IEEE1547AbnormalPerfomanceCategory,
    IEEE1547ControlSettings, IEEE1547Info, IEEE1547IslandingCategory,
    IEEE1547NormalPerformanceCategory, IEEE1547Setting, IEEE1547TripSettings,
    Impedance, IntegerQuantity, InterrupterUnitInfo, InterruptingMediumKind,
    InUseStateKind, IOPoint, IrregularIntervalSchedule, IrregularTimePoint,
    Jumper, Junction, KiloActivePower, Length, Limit, LimitSet, Line,
    LinearShuntCompensator, LinearShuntCompensatorPhase, LoadArea,
    LoadBreakSwitch, LoadGroup, LoadResponseCharacteristic, Location, Mass,
    Measurement, MeasurementValue, MeasurementValueQuality,
    MeasurementValueSource, Microgrid, Minutes, Money, MonthDay,
    MonthDayInterval, MutualCoupling, Name, NameType, NameTypeAuthority,
    NoLoadTest, NonConformLoad, NonConformLoadGroup, NonConformLoadSchedule,
    NonlinearShuntCompensator, NonlinearShuntCompensatorPhase,
    NonlinearShuntCompensatorPhasePoint, NonlinearShuntCompensatorPoint,
    OpenCircuitTest, OperatingMechanismInfo, OperatingMechanismKind,
    OperatingParticipant, OperatingShare, OperationalLimit,
    OperationalLimitDirectionKind, OperationalLimitSet, OperationalLimitType,
    OrderedPhaseCodeKind, OverfrequencyTripCurve, OverheadWireInfo,
    OvervoltageTripCurve, ParallelLineSegment, PerCent, PerLengthImpedance,
    PerLengthLineParameter, PerLengthPhaseImpedance,
    PerLengthSequenceImpedance, PetersenCoil, PetersenCoilModeKind, PhaseCode,
    PhaseImpedanceData, PhaseShuntConnectionKind, PhaseTapChanger,
    PhaseTapChangerAsymmetrical, PhaseTapChangerLinear,
    PhaseTapChangerNonLinear, PhaseTapChangerSymmetrical, PhaseTapChangerTable,
    PhaseTapChangerTablePoint, PhaseTapChangerTabular, PhotovoltaicUnit, Plant,
    PositionPoint, PowerCutZone, PowerElectronicsConnection,
    PowerElectronicsConnectionPhase, PowerElectronicsUnit,
    PowerElectronicsWindUnit, PowerSystemResource, PowerTransformer,
    PowerTransformerEnd, PowerTransformerInfo, Pressure, ProductAssetModel,
    ProtectedSwitch, ProtectionEquipment, ProtectionFunctionBlock,
    ProtectionSettingsGroup, PSRType, Quality61850, RaiseLowerCommand,
    RatioTapChanger, RatioTapChangerTable, RatioTapChangerTablePoint,
    Reactance, ReactancePerLength, ReactiveCapabilityCurve, ReactivePower,
    RealEnergy, Recloser, RegularIntervalSchedule, RegularTimePoint,
    RegulatingCondEq, RegulatingControl, RegulatingControlModeKind,
    RegulationSchedule, ReportingGroup, ReportingSuperGroup, Resistance,
    ResistancePerLength, RightOfWay, RotatingMachine, RotationSpeed,
    SchedulingArea, Season, SeasonDayTypeSchedule, SecondaryArea, Seconds,
    Sectionaliser, SeriesCompensator, SetPoint, ShortCircuitRotorKind,
    ShortCircuitTest, ShuntCompensator, ShuntCompensatorInfo,
    ShuntCompensatorPhase, SinglePhaseKind, SolarGeneratingUnit, Source,
    StaticVarCompensator, StationSupply, Status, StringMeasurement,
    StringMeasurementValue, StringQuantity, Structure, SubGeographicalRegion,
    SubLoadArea, SubSchedulingArea, Substation, Susceptance,
    SusceptancePerLength, SVCControlMode, SvShuntCompensatorSections,
    SvTapStep, Switch, SwitchArea, SwitchInfo, SwitchPhase, SwitchSchedule,
    SynchronousMachine, SynchronousMachineKind,
    SynchronousMachineOperatingMode, TapChanger, TapChangerControl,
    TapChangerInfo, TapChangerTablePoint, TapeShieldCableInfo, TapSchedule,
    Temperature, Terminal, ThermostatController, ThermostatControlMode,
    TimeInterval, TimeIntervalKind, TopologicalIsland, TopologicalNode,
    TransformerControlMode, TransformerCoreAdmittance, TransformerEnd,
    TransformerEndInfo, TransformerMeshImpedance, TransformerStarImpedance,
    TransformerTank, TransformerTankEnd, TransformerTankInfo, TransformerTest,
    UnderFrequencyProtectionFunctionBlock, UnderfrequencyTripCurve,
    UndervoltageTripCurve, UnitMultiplier, UnitSymbol, UsagePointGroup,
    Validity, ValueAliasSet, ValueToAlias, Version, Voltage,
    VoltageControlZone, VoltageLevel, VoltageLimit, VoltagePerReactivePower,
    VoltVarCurve, VoltWattCurve, Volume, WattVarCurve,
    WideAreaProtectionFunctionBlock, WindGeneratingUnit, WindGenUnitKind,
    WindingConnection, WireAssemblyInfo, WireInfo, WireInsulationKind,
    WireMaterialKind, WirePhaseInfo, WirePosition, WireSpacingInfo,
    WireUsageKind)

__all__ = [
    'Identity',
    'ACDCTerminal' ,
    'ACLineSegment' ,
    'ACLineSegmentPhase' ,
    'Accumulator' ,
    'AccumulatorLimit' ,
    'AccumulatorLimitSet' ,
    'AccumulatorReset' ,
    'AccumulatorValue' ,
    'ActivePowerLimit' ,
    'Analog' ,
    'AnalogControl' ,
    'AnalogLimit' ,
    'AnalogLimitSet' ,
    'AnalogValue' ,
    'ApparentPowerLimit' ,
    'AreaConfiguration' ,
    'Asset' ,
    'AssetContainer' ,
    'AssetInfo' ,
    'AssetKind' ,
    'AssetOrganisationRole' ,
    'AsynchronousMachine' ,
    'AsynchronousMachineKind' ,
    'BaseFrequency' ,
    'BasePower' ,
    'BaseVoltage' ,
    'BasicIntervalSchedule' ,
    'BatteryStateKind' ,
    'BatteryUnit' ,
    'Bay' ,
    'BranchGroup' ,
    'BranchGroupTerminal' ,
    'Breaker' ,
    'BreakerConfiguration' ,
    'BusNameMarker' ,
    'BusbarConfiguration' ,
    'BusbarSection' ,
    'BusbarSectionInfo' ,
    'Bushing' ,
    'BushingInfo' ,
    'BushingInsulationKind' ,
    'CableConstructionKind' ,
    'CableInfo' ,
    'CableOuterJacketKind' ,
    'CableShieldMaterialKind' ,
    'CatalogAssetType' ,
    'Clamp' ,
    'Command' ,
    'CompositeSwitch' ,
    'ConcentricNeutralCableInfo' ,
    'ConductingEquipment' ,
    'Conductor' ,
    'ConformLoad' ,
    'ConformLoadGroup' ,
    'ConformLoadSchedule' ,
    'ConnectivityNode' ,
    'ConnectivityNodeContainer' ,
    'Connector' ,
    'Control' ,
    'ConverterControlModeKind' ,
    'CoolantType' ,
    'CoordinateSystem' ,
    'Currency' ,
    'CurrentLimit' ,
    'Curve' ,
    'CurveData' ,
    'CurveStyle' ,
    'Customer' ,
    'CustomerAgreement' ,
    'CustomerKind' ,
    'Cut' ,
    'DERCurveData' ,
    'DERDynamics' ,
    'DERFunction' ,
    'DERGroupDispatch' ,
    'DERGroupForecast' ,
    'DERMonitorableParameter' ,
    'DERParameterKind' ,
    'DERUnitSymbol' ,
    'DateInterval' ,
    'DateTimeInterval' ,
    'DayType' ,
    'DecimalQuantity' ,
    'DemandResponseProgram' ,
    'Disconnector' ,
    'Discrete' ,
    'DiscreteValue' ,
    'DispatchSchedule' ,
    'DistributionArea' ,
    'DuctBank' ,
    'EarthFaultCompensator' ,
    'EndDevice' ,
    'EndDeviceControl' ,
    'EndDeviceGroup' ,
    'EnergyArea' ,
    'EnergyConnection' ,
    'EnergyConnectionProfile' ,
    'EnergyConsumer' ,
    'EnergyConsumerPhase' ,
    'EnergySource' ,
    'EnergySourcePhase' ,
    'Equipment' ,
    'EquipmentContainer' ,
    'ExternalNetworkInjection' ,
    'Feeder' ,
    'FeederArea' ,
    'FloatQuantity' ,
    'FlowDirectionKind' ,
    'FrequencyConverter' ,
    'FrequencyProtectionFunctionBlock' ,
    'FunctionBlock' ,
    'FunctionInputVariable' ,
    'FunctionOutputVariable' ,
    'Fuse' ,
    'GenUnitOpSchedule' ,
    'GeneratingUnit' ,
    'GeneratorControlMode' ,
    'GeneratorControlSource' ,
    'GeographicalRegion' ,
    'GrossToNetActivePowerCurve' ,
    'Ground' ,
    'GroundDisconnector' ,
    'GroundingImpedance' ,
    'House' ,
    'HouseCooling' ,
    'HouseHeating' ,
    'HouseThermalIntegrity' ,
    'HydroPump' ,
    'IEC61970CIMVersion' ,
    'IEEE1547AbnormalPerfomanceCategory' ,
    'IEEE1547ControlSettings' ,
    'IEEE1547Info' ,
    'IEEE1547IslandingCategory' ,
    'IEEE1547NormalPerformanceCategory' ,
    'IEEE1547Setting' ,
    'IEEE1547TripSettings' ,
    'IOPoint' ,
    'IdentifiedObject' ,
    'InUseStateKind' ,
    'IntegerQuantity' ,
    'InterrupterUnitInfo' ,
    'InterruptingMediumKind' ,
    'IrregularIntervalSchedule' ,
    'IrregularTimePoint' ,
    'Jumper' ,
    'Junction' ,
    'Limit' ,
    'LimitSet' ,
    'Line' ,
    'LinearShuntCompensator' ,
    'LinearShuntCompensatorPhase' ,
    'LoadArea' ,
    'LoadBreakSwitch' ,
    'LoadGroup' ,
    'LoadResponseCharacteristic' ,
    'Location' ,
    'Measurement' ,
    'MeasurementValue' ,
    'MeasurementValueQuality' ,
    'MeasurementValueSource' ,
    'Microgrid' ,
    'MonthDayInterval' ,
    'MutualCoupling' ,
    'Name' ,
    'NameType' ,
    'NameTypeAuthority' ,
    'NoLoadTest' ,
    'NonConformLoad' ,
    'NonConformLoadGroup' ,
    'NonConformLoadSchedule' ,
    'NonlinearShuntCompensator' ,
    'NonlinearShuntCompensatorPhase' ,
    'NonlinearShuntCompensatorPhasePoint' ,
    'NonlinearShuntCompensatorPoint' ,
    'OpenCircuitTest' ,
    'OperatingMechanismInfo' ,
    'OperatingMechanismKind' ,
    'OperatingParticipant' ,
    'OperatingShare' ,
    'OperationalLimit' ,
    'OperationalLimitDirectionKind' ,
    'OperationalLimitSet' ,
    'OperationalLimitType' ,
    'OrderedPhaseCodeKind' ,
    'OverfrequencyTripCurve' ,
    'OverheadWireInfo' ,
    'OvervoltageTripCurve' ,
    'PSRType' ,
    'ParallelLineSegment' ,
    'PerLengthImpedance' ,
    'PerLengthLineParameter' ,
    'PerLengthPhaseImpedance' ,
    'PerLengthSequenceImpedance' ,
    'PetersenCoil' ,
    'PetersenCoilModeKind' ,
    'PhaseCode' ,
    'PhaseImpedanceData' ,
    'PhaseShuntConnectionKind' ,
    'PhaseTapChanger' ,
    'PhaseTapChangerAsymmetrical' ,
    'PhaseTapChangerLinear' ,
    'PhaseTapChangerNonLinear' ,
    'PhaseTapChangerSymmetrical' ,
    'PhaseTapChangerTable' ,
    'PhaseTapChangerTablePoint' ,
    'PhaseTapChangerTabular' ,
    'PhotovoltaicUnit' ,
    'Plant' ,
    'PositionPoint' ,
    'PowerCutZone' ,
    'PowerElectronicsConnection' ,
    'PowerElectronicsConnectionPhase' ,
    'PowerElectronicsUnit' ,
    'PowerElectronicsWindUnit' ,
    'PowerSystemResource' ,
    'PowerTransformer' ,
    'PowerTransformerEnd' ,
    'PowerTransformerInfo' ,
    'ProductAssetModel' ,
    'ProtectedSwitch' ,
    'ProtectionEquipment' ,
    'ProtectionFunctionBlock' ,
    'ProtectionSettingsGroup' ,
    'Quality61850' ,
    'RaiseLowerCommand' ,
    'RatioTapChanger' ,
    'RatioTapChangerTable' ,
    'RatioTapChangerTablePoint' ,
    'ReactiveCapabilityCurve' ,
    'Recloser' ,
    'RegularIntervalSchedule' ,
    'RegularTimePoint' ,
    'RegulatingCondEq' ,
    'RegulatingControl' ,
    'RegulatingControlModeKind' ,
    'RegulationSchedule' ,
    'ReportingGroup' ,
    'ReportingSuperGroup' ,
    'RightOfWay' ,
    'RotatingMachine' ,
    'SVCControlMode' ,
    'SchedulingArea' ,
    'Season' ,
    'SeasonDayTypeSchedule' ,
    'SecondaryArea' ,
    'Sectionaliser' ,
    'SeriesCompensator' ,
    'SetPoint' ,
    'ShortCircuitRotorKind' ,
    'ShortCircuitTest' ,
    'ShuntCompensator' ,
    'ShuntCompensatorInfo' ,
    'ShuntCompensatorPhase' ,
    'SinglePhaseKind' ,
    'SolarGeneratingUnit' ,
    'StaticVarCompensator' ,
    'StationSupply' ,
    'Status' ,
    'StringMeasurement' ,
    'StringMeasurementValue' ,
    'StringQuantity' ,
    'Structure' ,
    'SubGeographicalRegion' ,
    'SubLoadArea' ,
    'SubSchedulingArea' ,
    'Substation' ,
    'SvShuntCompensatorSections' ,
    'SvTapStep' ,
    'Switch' ,
    'SwitchArea' ,
    'SwitchInfo' ,
    'SwitchPhase' ,
    'SwitchSchedule' ,
    'SynchronousMachine' ,
    'SynchronousMachineKind' ,
    'SynchronousMachineOperatingMode' ,
    'TapChanger' ,
    'TapChangerControl' ,
    'TapChangerInfo' ,
    'TapChangerTablePoint' ,
    'TapSchedule' ,
    'TapeShieldCableInfo' ,
    'Terminal' ,
    'ThermostatControlMode' ,
    'ThermostatController' ,
    'TimeInterval' ,
    'TimeIntervalKind' ,
    'TopologicalIsland' ,
    'TopologicalNode' ,
    'TransformerControlMode' ,
    'TransformerCoreAdmittance' ,
    'TransformerEnd' ,
    'TransformerEndInfo' ,
    'TransformerMeshImpedance' ,
    'TransformerStarImpedance' ,
    'TransformerTank' ,
    'TransformerTankEnd' ,
    'TransformerTankInfo' ,
    'TransformerTest' ,
    'UnderFrequencyProtectionFunctionBlock' ,
    'UnderfrequencyTripCurve' ,
    'UndervoltageTripCurve' ,
    'UnitMultiplier' ,
    'UnitSymbol' ,
    'UsagePointGroup' ,
    'Validity' ,
    'ValueAliasSet' ,
    'ValueToAlias' ,
    'Version' ,
    'VoltVarCurve' ,
    'VoltWattCurve' ,
    'VoltageControlZone' ,
    'VoltageLevel' ,
    'VoltageLimit' ,
    'WattVarCurve' ,
    'WideAreaProtectionFunctionBlock' ,
    'WindGenUnitKind' ,
    'WindGeneratingUnit' ,
    'WindingConnection' ,
    'WireAssemblyInfo' ,
    'WireInfo' ,
    'WireInsulationKind' ,
    'WireMaterialKind' ,
    'WirePhaseInfo' ,
    'WirePosition' ,
    'WireSpacingInfo' ,
    'WireUsageKind' ,
    'MonthDay' ,
    'Susceptance' ,
    'ReactivePower' ,
    'Pressure' ,
    'Capacitance' ,
    'Temperature' ,
    'Seconds' ,
    'Voltage' ,
    'ActivePowerPerFrequency' ,
    'SusceptancePerLength' ,
    'Frequency' ,
    'PU' ,
    'VoltagePerReactivePower' ,
    'ActivePower' ,
    'Mass' ,
    'ConductancePerLength' ,
    'Reactance' ,
    'Minutes' ,
    'AngleDegrees' ,
    'RotationSpeed' ,
    'RealEnergy' ,
    'Conductance' ,
    'ReactancePerLength' ,
    'Resistance' ,
    'Money' ,
    'Area' ,
    'Volume' ,
    'Length' ,
    'KiloActivePower' ,
    'ResistancePerLength' ,
    'Displacement' ,
    'ApparentPower' ,
    'PerCent' ,
    'AngleRadians' ,
    'CurrentFlow' ,
    'Source' ,
    'Impedance'
]
